#ifndef _stdint_h__
#define _stdint_h__

#include "../../bundled/zeromq/src/stdint.hpp"

#define UINT8_MAX 0xff
#define UINT16_MAX 0xffff
#define UINT32_MAX 0xffffffff
#define UINT64_MAX 0xffffffffffffffffull

#endif /* _stdint_h__ */
