/* Generated by Cython 0.24 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__zmq__devices__monitoredqueue
#define __PYX_HAVE_API__zmq__devices__monitoredqueue
#include "zmq_compat.h"
#include "zmq.h"
#include "zmq_utils.h"
#include "string.h"
#include "pyversion_compat.h"
#include "errno.h"
#include "stdio.h"
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "zmq/devices/monitoredqueue.pyx",
  "zmq/utils/buffers.pxd",
  "zmq/backend/cython/checkrc.pxd",
  "zmq/backend/cython/context.pxd",
  "zmq/backend/cython/socket.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_3zmq_7backend_6cython_7context_Context;
struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket;
struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_send;
struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_recv;

/* "zmq/backend/cython/socket.pxd":45
 * 
 *     # cpdef methods for direct-cython access:
 *     cpdef object send(self, object data, int flags=*, copy=*, track=*)             # <<<<<<<<<<<<<<
 *     cpdef object recv(self, int flags=*, copy=*, track=*)
 * 
 */
struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_send {
  int __pyx_n;
  int flags;
  PyObject *copy;
  PyObject *track;
};

/* "zmq/backend/cython/socket.pxd":46
 *     # cpdef methods for direct-cython access:
 *     cpdef object send(self, object data, int flags=*, copy=*, track=*)
 *     cpdef object recv(self, int flags=*, copy=*, track=*)             # <<<<<<<<<<<<<<
 * 
 */
struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_recv {
  int __pyx_n;
  int flags;
  PyObject *copy;
  PyObject *track;
};

/* "context.pxd":26
 * #-----------------------------------------------------------------------------
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 * 
 *     cdef object __weakref__     # enable weakref
 */
struct __pyx_obj_3zmq_7backend_6cython_7context_Context {
  PyObject_HEAD
  struct __pyx_vtabstruct_3zmq_7backend_6cython_7context_Context *__pyx_vtab;
  PyObject *__weakref__;
  void *handle;
  int _shadow;
  void **_sockets;
  size_t _n_sockets;
  size_t _max_sockets;
  int _pid;
  int closed;
};


/* "zmq/backend/cython/socket.pxd":33
 * 
 * 
 * cdef class Socket:             # <<<<<<<<<<<<<<
 * 
 *     cdef object __weakref__     # enable weakref
 */
struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket {
  PyObject_HEAD
  struct __pyx_vtabstruct_3zmq_7backend_6cython_6socket_Socket *__pyx_vtab;
  PyObject *__weakref__;
  void *handle;
  int _shadow;
  struct __pyx_obj_3zmq_7backend_6cython_7context_Context *context;
  int _closed;
  int _pid;
};



/* "context.pxd":26
 * #-----------------------------------------------------------------------------
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 * 
 *     cdef object __weakref__     # enable weakref
 */

struct __pyx_vtabstruct_3zmq_7backend_6cython_7context_Context {
  int (*_term)(struct __pyx_obj_3zmq_7backend_6cython_7context_Context *);
  void (*_add_socket)(struct __pyx_obj_3zmq_7backend_6cython_7context_Context *, void *);
  void (*_remove_socket)(struct __pyx_obj_3zmq_7backend_6cython_7context_Context *, void *);
};
static struct __pyx_vtabstruct_3zmq_7backend_6cython_7context_Context *__pyx_vtabptr_3zmq_7backend_6cython_7context_Context;


/* "zmq/backend/cython/socket.pxd":33
 * 
 * 
 * cdef class Socket:             # <<<<<<<<<<<<<<
 * 
 *     cdef object __weakref__     # enable weakref
 */

struct __pyx_vtabstruct_3zmq_7backend_6cython_6socket_Socket {
  PyObject *(*send)(struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_send *__pyx_optional_args);
  PyObject *(*recv)(struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *, int __pyx_skip_dispatch, struct __pyx_opt_args_3zmq_7backend_6cython_6socket_6Socket_recv *__pyx_optional_args);
};
static struct __pyx_vtabstruct_3zmq_7backend_6cython_6socket_Socket *__pyx_vtabptr_3zmq_7backend_6cython_6socket_Socket;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* SaveResetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libzmq' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'buffers' */
static CYTHON_INLINE int __pyx_f_7buffers_memoryview_available(void); /*proto*/
static CYTHON_INLINE int __pyx_f_7buffers_oldstyle_available(void); /*proto*/
static CYTHON_INLINE int __pyx_f_7buffers_check_buffer(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_asbuffer(PyObject *, int, int, void **, Py_ssize_t *, Py_ssize_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_asbuffer_r(PyObject *, void **, Py_ssize_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_3(void *, Py_ssize_t, int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_2(void *, Py_ssize_t, int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer(void *, Py_ssize_t, int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7buffers_viewfromobject(PyObject *, int); /*proto*/

/* Module declarations from 'zmq.backend.cython.context' */
static PyTypeObject *__pyx_ptype_3zmq_7backend_6cython_7context_Context = 0;

/* Module declarations from 'zmq.backend.cython.socket' */
static PyTypeObject *__pyx_ptype_3zmq_7backend_6cython_6socket_Socket = 0;

/* Module declarations from 'libc.errno' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'zmq.backend.cython.libzmq' */

/* Module declarations from 'zmq.backend.cython.checkrc' */
static CYTHON_INLINE int __pyx_f_3zmq_7backend_6cython_7checkrc__check_rc(int); /*proto*/

/* Module declarations from 'zmq.devices.monitoredqueue' */
static CYTHON_INLINE int __pyx_f_3zmq_7devices_14monitoredqueue__relay(void *, void *, void *, zmq_msg_t, zmq_msg_t, zmq_msg_t, int); /*proto*/
static CYTHON_INLINE int __pyx_f_3zmq_7devices_14monitoredqueue_c_monitored_queue(void *, void *, void *, zmq_msg_t *, zmq_msg_t *, int); /*proto*/
#define __Pyx_MODULE_NAME "zmq.devices.monitoredqueue"
int __pyx_module_is_main_zmq__devices__monitoredqueue = 0;

/* Implementation of 'zmq.devices.monitoredqueue' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_NotImplementedError;
static const char __pyx_k_B[] = "B";
static const char __pyx_k__2[] = "";
static const char __pyx_k_in[] = "in";
static const char __pyx_k_rc[] = "rc";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_ins[] = "ins";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_zmq[] = "zmq";
static const char __pyx_k_char[] = "char";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mons[] = "mons";
static const char __pyx_k_outs[] = "outs";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_Again[] = "Again";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_msg_c[] = "msg_c";
static const char __pyx_k_ROUTER[] = "ROUTER";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_in_msg[] = "in_msg";
static const char __pyx_k_out_msg[] = "out_msg";
static const char __pyx_k_ZMQError[] = "ZMQError";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_swap_ids[] = "swap_ids";
static const char __pyx_k_typecode[] = "typecode";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_in_prefix[] = "in_prefix";
static const char __pyx_k_in_socket[] = "in_socket";
static const char __pyx_k_msg_c_len[] = "msg_c_len";
static const char __pyx_k_zmq_error[] = "zmq.error";
static const char __pyx_k_mon_socket[] = "mon_socket";
static const char __pyx_k_out_prefix[] = "out_prefix";
static const char __pyx_k_out_socket[] = "out_socket";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_monitored_queue[] = "monitored_queue";
static const char __pyx_k_ContextTerminated[] = "ContextTerminated";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_InterruptedSystemCall[] = "InterruptedSystemCall";
static const char __pyx_k_zmq_devices_monitoredqueue[] = "zmq.devices.monitoredqueue";
static const char __pyx_k_Old_style_buffers_not_available[] = "Old style buffers not available.";
static const char __pyx_k_private_tmp_pyzmq_release_zmq_d[] = "/private/tmp/pyzmq-release/zmq/devices/monitoredqueue.pyx";
static const char __pyx_k_r_does_not_provide_a_buffer_int[] = "%r does not provide a buffer interface.";
static const char __pyx_k_MonitoredQueue_classes_and_funct[] = "MonitoredQueue classes and functions.\n\nAuthors\n-------\n* MinRK\n* Brian Granger\n";
static PyObject *__pyx_n_s_Again;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_b_B;
static PyObject *__pyx_n_s_ContextTerminated;
static PyObject *__pyx_n_s_InterruptedSystemCall;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_kp_s_Old_style_buffers_not_available;
static PyObject *__pyx_n_s_ROUTER;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ZMQError;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_char;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_b_in;
static PyObject *__pyx_n_s_in_msg;
static PyObject *__pyx_n_s_in_prefix;
static PyObject *__pyx_n_s_in_socket;
static PyObject *__pyx_n_s_ins;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mon_socket;
static PyObject *__pyx_n_s_monitored_queue;
static PyObject *__pyx_n_s_mons;
static PyObject *__pyx_n_s_msg_c;
static PyObject *__pyx_n_s_msg_c_len;
static PyObject *__pyx_n_b_out;
static PyObject *__pyx_n_s_out_msg;
static PyObject *__pyx_n_s_out_prefix;
static PyObject *__pyx_n_s_out_socket;
static PyObject *__pyx_n_s_outs;
static PyObject *__pyx_kp_s_private_tmp_pyzmq_release_zmq_d;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_r_does_not_provide_a_buffer_int;
static PyObject *__pyx_n_s_rc;
static PyObject *__pyx_n_s_swap_ids;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_typecode;
static PyObject *__pyx_n_s_zmq;
static PyObject *__pyx_n_s_zmq_devices_monitoredqueue;
static PyObject *__pyx_n_s_zmq_error;
static PyObject *__pyx_pf_3zmq_7devices_14monitoredqueue_monitored_queue(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_in_socket, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_out_socket, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_mon_socket, PyObject *__pyx_v_in_prefix, PyObject *__pyx_v_out_prefix); /* proto */
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_codeobj__5;

/* "zmq/devices/monitoredqueue.pyx":41
 * 
 * 
 * def monitored_queue(Socket in_socket, Socket out_socket, Socket mon_socket,             # <<<<<<<<<<<<<<
 *                     bytes in_prefix=b'in', bytes out_prefix=b'out'):
 *     """monitored_queue(in_socket, out_socket, mon_socket,
 */

/* Python wrapper */
static PyObject *__pyx_pw_3zmq_7devices_14monitoredqueue_1monitored_queue(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3zmq_7devices_14monitoredqueue_monitored_queue[] = "monitored_queue(in_socket, out_socket, mon_socket,\n                       in_prefix=b'in', out_prefix=b'out')\n    \n    Start a monitored queue device.\n    \n    A monitored queue is very similar to the zmq.proxy device (monitored queue came first).\n    \n    Differences from zmq.proxy:\n    \n    - monitored_queue supports both in and out being ROUTER sockets\n      (via swapping IDENTITY prefixes).\n    - monitor messages are prefixed, making in and out messages distinguishable.\n    \n    Parameters\n    ----------\n    in_socket : Socket\n        One of the sockets to the Queue. Its messages will be prefixed with\n        'in'.\n    out_socket : Socket\n        One of the sockets to the Queue. Its messages will be prefixed with\n        'out'. The only difference between in/out socket is this prefix.\n    mon_socket : Socket\n        This socket sends out every message received by each of the others\n        with an in/out prefix specifying which one it was.\n    in_prefix : str\n        Prefix added to broadcast messages from in_socket.\n    out_prefix : str\n        Prefix added to broadcast messages from out_socket.\n    ";
static PyMethodDef __pyx_mdef_3zmq_7devices_14monitoredqueue_1monitored_queue = {"monitored_queue", (PyCFunction)__pyx_pw_3zmq_7devices_14monitoredqueue_1monitored_queue, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3zmq_7devices_14monitoredqueue_monitored_queue};
static PyObject *__pyx_pw_3zmq_7devices_14monitoredqueue_1monitored_queue(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_in_socket = 0;
  struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_out_socket = 0;
  struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_mon_socket = 0;
  PyObject *__pyx_v_in_prefix = 0;
  PyObject *__pyx_v_out_prefix = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("monitored_queue (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_in_socket,&__pyx_n_s_out_socket,&__pyx_n_s_mon_socket,&__pyx_n_s_in_prefix,&__pyx_n_s_out_prefix,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject*)__pyx_n_b_in);
    values[4] = ((PyObject*)__pyx_n_b_out);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_in_socket)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_out_socket)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("monitored_queue", 0, 3, 5, 1); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mon_socket)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("monitored_queue", 0, 3, 5, 2); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_in_prefix);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_out_prefix);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "monitored_queue") < 0)) __PYX_ERR(0, 41, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_in_socket = ((struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *)values[0]);
    __pyx_v_out_socket = ((struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *)values[1]);
    __pyx_v_mon_socket = ((struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *)values[2]);
    __pyx_v_in_prefix = ((PyObject*)values[3]);
    __pyx_v_out_prefix = ((PyObject*)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("monitored_queue", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 41, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("zmq.devices.monitoredqueue.monitored_queue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_in_socket), __pyx_ptype_3zmq_7backend_6cython_6socket_Socket, 1, "in_socket", 0))) __PYX_ERR(0, 41, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_out_socket), __pyx_ptype_3zmq_7backend_6cython_6socket_Socket, 1, "out_socket", 0))) __PYX_ERR(0, 41, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mon_socket), __pyx_ptype_3zmq_7backend_6cython_6socket_Socket, 1, "mon_socket", 0))) __PYX_ERR(0, 41, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_in_prefix), (&PyBytes_Type), 1, "in_prefix", 1))) __PYX_ERR(0, 42, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_out_prefix), (&PyBytes_Type), 1, "out_prefix", 1))) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_r = __pyx_pf_3zmq_7devices_14monitoredqueue_monitored_queue(__pyx_self, __pyx_v_in_socket, __pyx_v_out_socket, __pyx_v_mon_socket, __pyx_v_in_prefix, __pyx_v_out_prefix);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3zmq_7devices_14monitoredqueue_monitored_queue(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_in_socket, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_out_socket, struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket *__pyx_v_mon_socket, PyObject *__pyx_v_in_prefix, PyObject *__pyx_v_out_prefix) {
  void *__pyx_v_ins;
  void *__pyx_v_outs;
  void *__pyx_v_mons;
  zmq_msg_t __pyx_v_in_msg;
  zmq_msg_t __pyx_v_out_msg;
  int __pyx_v_swap_ids;
  char *__pyx_v_msg_c;
  Py_ssize_t __pyx_v_msg_c_len;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("monitored_queue", 0);

  /* "zmq/devices/monitoredqueue.pyx":73
 *     """
 * 
 *     cdef void *ins=in_socket.handle             # <<<<<<<<<<<<<<
 *     cdef void *outs=out_socket.handle
 *     cdef void *mons=mon_socket.handle
 */
  __pyx_t_1 = __pyx_v_in_socket->handle;
  __pyx_v_ins = __pyx_t_1;

  /* "zmq/devices/monitoredqueue.pyx":74
 * 
 *     cdef void *ins=in_socket.handle
 *     cdef void *outs=out_socket.handle             # <<<<<<<<<<<<<<
 *     cdef void *mons=mon_socket.handle
 *     cdef zmq_msg_t in_msg
 */
  __pyx_t_1 = __pyx_v_out_socket->handle;
  __pyx_v_outs = __pyx_t_1;

  /* "zmq/devices/monitoredqueue.pyx":75
 *     cdef void *ins=in_socket.handle
 *     cdef void *outs=out_socket.handle
 *     cdef void *mons=mon_socket.handle             # <<<<<<<<<<<<<<
 *     cdef zmq_msg_t in_msg
 *     cdef zmq_msg_t out_msg
 */
  __pyx_t_1 = __pyx_v_mon_socket->handle;
  __pyx_v_mons = __pyx_t_1;

  /* "zmq/devices/monitoredqueue.pyx":79
 *     cdef zmq_msg_t out_msg
 *     cdef bint swap_ids
 *     cdef char *msg_c = NULL             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t msg_c_len
 *     cdef int rc
 */
  __pyx_v_msg_c = NULL;

  /* "zmq/devices/monitoredqueue.pyx":84
 * 
 *     # force swap_ids if both ROUTERs
 *     swap_ids = (in_socket.type == ROUTER and out_socket.type == ROUTER)             # <<<<<<<<<<<<<<
 * 
 *     # build zmq_msg objects from str prefixes
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_in_socket), __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ROUTER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_out_socket), __pyx_n_s_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ROUTER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  __pyx_v_swap_ids = __pyx_t_2;

  /* "zmq/devices/monitoredqueue.pyx":87
 * 
 *     # build zmq_msg objects from str prefixes
 *     asbuffer_r(in_prefix, <void **>&msg_c, &msg_c_len)             # <<<<<<<<<<<<<<
 *     rc = zmq_msg_init_size(&in_msg, msg_c_len)
 *     _check_rc(rc)
 */
  __pyx_t_3 = __pyx_f_7buffers_asbuffer_r(__pyx_v_in_prefix, ((void **)(&__pyx_v_msg_c)), (&__pyx_v_msg_c_len)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "zmq/devices/monitoredqueue.pyx":88
 *     # build zmq_msg objects from str prefixes
 *     asbuffer_r(in_prefix, <void **>&msg_c, &msg_c_len)
 *     rc = zmq_msg_init_size(&in_msg, msg_c_len)             # <<<<<<<<<<<<<<
 *     _check_rc(rc)
 * 
 */
  __pyx_v_rc = zmq_msg_init_size((&__pyx_v_in_msg), __pyx_v_msg_c_len);

  /* "zmq/devices/monitoredqueue.pyx":89
 *     asbuffer_r(in_prefix, <void **>&msg_c, &msg_c_len)
 *     rc = zmq_msg_init_size(&in_msg, msg_c_len)
 *     _check_rc(rc)             # <<<<<<<<<<<<<<
 * 
 *     memcpy(zmq_msg_data(&in_msg), msg_c, zmq_msg_size(&in_msg))
 */
  __pyx_t_7 = __pyx_f_3zmq_7backend_6cython_7checkrc__check_rc(__pyx_v_rc); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 89, __pyx_L1_error)

  /* "zmq/devices/monitoredqueue.pyx":91
 *     _check_rc(rc)
 * 
 *     memcpy(zmq_msg_data(&in_msg), msg_c, zmq_msg_size(&in_msg))             # <<<<<<<<<<<<<<
 * 
 *     asbuffer_r(out_prefix, <void **>&msg_c, &msg_c_len)
 */
  memcpy(zmq_msg_data((&__pyx_v_in_msg)), __pyx_v_msg_c, zmq_msg_size((&__pyx_v_in_msg)));

  /* "zmq/devices/monitoredqueue.pyx":93
 *     memcpy(zmq_msg_data(&in_msg), msg_c, zmq_msg_size(&in_msg))
 * 
 *     asbuffer_r(out_prefix, <void **>&msg_c, &msg_c_len)             # <<<<<<<<<<<<<<
 * 
 *     rc = zmq_msg_init_size(&out_msg, msg_c_len)
 */
  __pyx_t_3 = __pyx_f_7buffers_asbuffer_r(__pyx_v_out_prefix, ((void **)(&__pyx_v_msg_c)), (&__pyx_v_msg_c_len)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "zmq/devices/monitoredqueue.pyx":95
 *     asbuffer_r(out_prefix, <void **>&msg_c, &msg_c_len)
 * 
 *     rc = zmq_msg_init_size(&out_msg, msg_c_len)             # <<<<<<<<<<<<<<
 *     _check_rc(rc)
 * 
 */
  __pyx_v_rc = zmq_msg_init_size((&__pyx_v_out_msg), __pyx_v_msg_c_len);

  /* "zmq/devices/monitoredqueue.pyx":96
 * 
 *     rc = zmq_msg_init_size(&out_msg, msg_c_len)
 *     _check_rc(rc)             # <<<<<<<<<<<<<<
 * 
 *     while True:
 */
  __pyx_t_7 = __pyx_f_3zmq_7backend_6cython_7checkrc__check_rc(__pyx_v_rc); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 96, __pyx_L1_error)

  /* "zmq/devices/monitoredqueue.pyx":98
 *     _check_rc(rc)
 * 
 *     while True:             # <<<<<<<<<<<<<<
 *         with nogil:
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 */
  while (1) {

    /* "zmq/devices/monitoredqueue.pyx":99
 * 
 *     while True:
 *         with nogil:             # <<<<<<<<<<<<<<
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "zmq/devices/monitoredqueue.pyx":100
 *     while True:
 *         with nogil:
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))             # <<<<<<<<<<<<<<
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 *         try:
 */
          memcpy(zmq_msg_data((&__pyx_v_out_msg)), __pyx_v_msg_c, zmq_msg_size((&__pyx_v_out_msg)));

          /* "zmq/devices/monitoredqueue.pyx":101
 *         with nogil:
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)             # <<<<<<<<<<<<<<
 *         try:
 *             _check_rc(rc)
 */
          __pyx_v_rc = __pyx_f_3zmq_7devices_14monitoredqueue_c_monitored_queue(__pyx_v_ins, __pyx_v_outs, __pyx_v_mons, (&__pyx_v_in_msg), (&__pyx_v_out_msg), __pyx_v_swap_ids);
        }

        /* "zmq/devices/monitoredqueue.pyx":99
 * 
 *     while True:
 *         with nogil:             # <<<<<<<<<<<<<<
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L11;
          }
          __pyx_L11:;
        }
    }

    /* "zmq/devices/monitoredqueue.pyx":102
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 *         try:             # <<<<<<<<<<<<<<
 *             _check_rc(rc)
 *         except InterruptedSystemCall:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "zmq/devices/monitoredqueue.pyx":103
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 *         try:
 *             _check_rc(rc)             # <<<<<<<<<<<<<<
 *         except InterruptedSystemCall:
 *             continue
 */
        __pyx_t_7 = __pyx_f_3zmq_7backend_6cython_7checkrc__check_rc(__pyx_v_rc); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 103, __pyx_L12_error)

        /* "zmq/devices/monitoredqueue.pyx":102
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 *         try:             # <<<<<<<<<<<<<<
 *             _check_rc(rc)
 *         except InterruptedSystemCall:
 */
      }

      /* "zmq/devices/monitoredqueue.pyx":107
 *             continue
 *         else:
 *             break             # <<<<<<<<<<<<<<
 *     return rc
 * 
 */
      /*else:*/ {
        goto __pyx_L17_try_break;
      }
      __pyx_L12_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "zmq/devices/monitoredqueue.pyx":104
 *         try:
 *             _check_rc(rc)
 *         except InterruptedSystemCall:             # <<<<<<<<<<<<<<
 *             continue
 *         else:
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_InterruptedSystemCall); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 104, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        __Pyx_AddTraceback("zmq.devices.monitoredqueue.monitored_queue", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 104, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_5);

        /* "zmq/devices/monitoredqueue.pyx":105
 *             _check_rc(rc)
 *         except InterruptedSystemCall:
 *             continue             # <<<<<<<<<<<<<<
 *         else:
 *             break
 */
        goto __pyx_L21_except_continue;
        __pyx_L21_except_continue:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L18_try_continue;
      }
      goto __pyx_L14_except_error;
      __pyx_L14_except_error:;

      /* "zmq/devices/monitoredqueue.pyx":102
 *             memcpy(zmq_msg_data(&out_msg), msg_c, zmq_msg_size(&out_msg))
 *             rc = c_monitored_queue(ins, outs, mons, &in_msg, &out_msg, swap_ids)
 *         try:             # <<<<<<<<<<<<<<
 *             _check_rc(rc)
 *         except InterruptedSystemCall:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L17_try_break:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L6_break;
      __pyx_L18_try_continue:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L5_continue;
    }
    __pyx_L5_continue:;
  }
  __pyx_L6_break:;

  /* "zmq/devices/monitoredqueue.pyx":108
 *         else:
 *             break
 *     return rc             # <<<<<<<<<<<<<<
 * 
 * __all__ = ['monitored_queue']
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "zmq/devices/monitoredqueue.pyx":41
 * 
 * 
 * def monitored_queue(Socket in_socket, Socket out_socket, Socket mon_socket,             # <<<<<<<<<<<<<<
 *                     bytes in_prefix=b'in', bytes out_prefix=b'out'):
 *     """monitored_queue(in_socket, out_socket, mon_socket,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("zmq.devices.monitoredqueue.monitored_queue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":88
 * 
 * 
 * cdef inline int memoryview_available():             # <<<<<<<<<<<<<<
 *     return PY_MAJOR_VERSION >= 3 or (PY_MAJOR_VERSION >=2 and PY_MINOR_VERSION >= 7)
 * 
 */

static CYTHON_INLINE int __pyx_f_7buffers_memoryview_available(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("memoryview_available", 0);

  /* "buffers.pxd":89
 * 
 * cdef inline int memoryview_available():
 *     return PY_MAJOR_VERSION >= 3 or (PY_MAJOR_VERSION >=2 and PY_MINOR_VERSION >= 7)             # <<<<<<<<<<<<<<
 * 
 * cdef inline int oldstyle_available():
 */
  __pyx_t_2 = (PY_MAJOR_VERSION >= 3);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (PY_MAJOR_VERSION >= 2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (PY_MINOR_VERSION >= 7);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "buffers.pxd":88
 * 
 * 
 * cdef inline int memoryview_available():             # <<<<<<<<<<<<<<
 *     return PY_MAJOR_VERSION >= 3 or (PY_MAJOR_VERSION >=2 and PY_MINOR_VERSION >= 7)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":91
 *     return PY_MAJOR_VERSION >= 3 or (PY_MAJOR_VERSION >=2 and PY_MINOR_VERSION >= 7)
 * 
 * cdef inline int oldstyle_available():             # <<<<<<<<<<<<<<
 *     return PY_MAJOR_VERSION < 3
 * 
 */

static CYTHON_INLINE int __pyx_f_7buffers_oldstyle_available(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("oldstyle_available", 0);

  /* "buffers.pxd":92
 * 
 * cdef inline int oldstyle_available():
 *     return PY_MAJOR_VERSION < 3             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (PY_MAJOR_VERSION < 3);
  goto __pyx_L0;

  /* "buffers.pxd":91
 *     return PY_MAJOR_VERSION >= 3 or (PY_MAJOR_VERSION >=2 and PY_MINOR_VERSION >= 7)
 * 
 * cdef inline int oldstyle_available():             # <<<<<<<<<<<<<<
 *     return PY_MAJOR_VERSION < 3
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":95
 * 
 * 
 * cdef inline int check_buffer(object ob):             # <<<<<<<<<<<<<<
 *     """Version independent check for whether an object is a buffer.
 * 
 */

static CYTHON_INLINE int __pyx_f_7buffers_check_buffer(PyObject *__pyx_v_ob) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("check_buffer", 0);

  /* "buffers.pxd":107
 *     int : 0 if no buffer interface, 3 if newstyle buffer interface, 2 if oldstyle.
 *     """
 *     if PyObject_CheckBuffer(ob):             # <<<<<<<<<<<<<<
 *         return 3
 *     if oldstyle_available():
 */
  __pyx_t_1 = (PyObject_CheckBuffer(__pyx_v_ob) != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":108
 *     """
 *     if PyObject_CheckBuffer(ob):
 *         return 3             # <<<<<<<<<<<<<<
 *     if oldstyle_available():
 *         return PyObject_CheckReadBuffer(ob) and 2
 */
    __pyx_r = 3;
    goto __pyx_L0;

    /* "buffers.pxd":107
 *     int : 0 if no buffer interface, 3 if newstyle buffer interface, 2 if oldstyle.
 *     """
 *     if PyObject_CheckBuffer(ob):             # <<<<<<<<<<<<<<
 *         return 3
 *     if oldstyle_available():
 */
  }

  /* "buffers.pxd":109
 *     if PyObject_CheckBuffer(ob):
 *         return 3
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         return PyObject_CheckReadBuffer(ob) and 2
 *     return 0
 */
  __pyx_t_1 = (__pyx_f_7buffers_oldstyle_available() != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":110
 *         return 3
 *     if oldstyle_available():
 *         return PyObject_CheckReadBuffer(ob) and 2             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_t_3 = PyObject_CheckReadBuffer(__pyx_v_ob);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = 2;
    __pyx_L5_bool_binop_done:;
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "buffers.pxd":109
 *     if PyObject_CheckBuffer(ob):
 *         return 3
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         return PyObject_CheckReadBuffer(ob) and 2
 *     return 0
 */
  }

  /* "buffers.pxd":111
 *     if oldstyle_available():
 *         return PyObject_CheckReadBuffer(ob) and 2
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "buffers.pxd":95
 * 
 * 
 * cdef inline int check_buffer(object ob):             # <<<<<<<<<<<<<<
 *     """Version independent check for whether an object is a buffer.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":114
 * 
 * 
 * cdef inline object asbuffer(object ob, int writable, int format,             # <<<<<<<<<<<<<<
 *                             void **base, Py_ssize_t *size,
 *                             Py_ssize_t *itemsize):
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_asbuffer(PyObject *__pyx_v_ob, int __pyx_v_writable, int __pyx_v_format, void **__pyx_v_base, Py_ssize_t *__pyx_v_size, Py_ssize_t *__pyx_v_itemsize) {
  void *__pyx_v_bptr;
  Py_ssize_t __pyx_v_blen;
  Py_ssize_t __pyx_v_bitemlen;
  Py_buffer __pyx_v_view;
  int __pyx_v_flags;
  int __pyx_v_mode;
  PyObject *__pyx_v_bfmt = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  void *__pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("asbuffer", 0);

  /* "buffers.pxd":141
 *     """
 * 
 *     cdef void *bptr = NULL             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t blen = 0, bitemlen = 0
 *     cdef Py_buffer view
 */
  __pyx_v_bptr = NULL;

  /* "buffers.pxd":142
 * 
 *     cdef void *bptr = NULL
 *     cdef Py_ssize_t blen = 0, bitemlen = 0             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef int flags = PyBUF_SIMPLE
 */
  __pyx_v_blen = 0;
  __pyx_v_bitemlen = 0;

  /* "buffers.pxd":144
 *     cdef Py_ssize_t blen = 0, bitemlen = 0
 *     cdef Py_buffer view
 *     cdef int flags = PyBUF_SIMPLE             # <<<<<<<<<<<<<<
 *     cdef int mode = 0
 * 
 */
  __pyx_v_flags = PyBUF_SIMPLE;

  /* "buffers.pxd":145
 *     cdef Py_buffer view
 *     cdef int flags = PyBUF_SIMPLE
 *     cdef int mode = 0             # <<<<<<<<<<<<<<
 * 
 *     bfmt = None
 */
  __pyx_v_mode = 0;

  /* "buffers.pxd":147
 *     cdef int mode = 0
 * 
 *     bfmt = None             # <<<<<<<<<<<<<<
 * 
 *     mode = check_buffer(ob)
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_bfmt = Py_None;

  /* "buffers.pxd":149
 *     bfmt = None
 * 
 *     mode = check_buffer(ob)             # <<<<<<<<<<<<<<
 *     if mode == 0:
 *         raise TypeError("%r does not provide a buffer interface."%ob)
 */
  __pyx_v_mode = __pyx_f_7buffers_check_buffer(__pyx_v_ob);

  /* "buffers.pxd":150
 * 
 *     mode = check_buffer(ob)
 *     if mode == 0:             # <<<<<<<<<<<<<<
 *         raise TypeError("%r does not provide a buffer interface."%ob)
 * 
 */
  __pyx_t_1 = ((__pyx_v_mode == 0) != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":151
 *     mode = check_buffer(ob)
 *     if mode == 0:
 *         raise TypeError("%r does not provide a buffer interface."%ob)             # <<<<<<<<<<<<<<
 * 
 *     if mode == 3:
 */
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_r_does_not_provide_a_buffer_int, __pyx_v_ob); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 151, __pyx_L1_error)

    /* "buffers.pxd":150
 * 
 *     mode = check_buffer(ob)
 *     if mode == 0:             # <<<<<<<<<<<<<<
 *         raise TypeError("%r does not provide a buffer interface."%ob)
 * 
 */
  }

  /* "buffers.pxd":153
 *         raise TypeError("%r does not provide a buffer interface."%ob)
 * 
 *     if mode == 3:             # <<<<<<<<<<<<<<
 *         flags = PyBUF_ANY_CONTIGUOUS
 *         if writable:
 */
  __pyx_t_1 = ((__pyx_v_mode == 3) != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":154
 * 
 *     if mode == 3:
 *         flags = PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if writable:
 *             flags |= PyBUF_WRITABLE
 */
    __pyx_v_flags = PyBUF_ANY_CONTIGUOUS;

    /* "buffers.pxd":155
 *     if mode == 3:
 *         flags = PyBUF_ANY_CONTIGUOUS
 *         if writable:             # <<<<<<<<<<<<<<
 *             flags |= PyBUF_WRITABLE
 *         if format:
 */
    __pyx_t_1 = (__pyx_v_writable != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":156
 *         flags = PyBUF_ANY_CONTIGUOUS
 *         if writable:
 *             flags |= PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         if format:
 *             flags |= PyBUF_FORMAT
 */
      __pyx_v_flags = (__pyx_v_flags | PyBUF_WRITABLE);

      /* "buffers.pxd":155
 *     if mode == 3:
 *         flags = PyBUF_ANY_CONTIGUOUS
 *         if writable:             # <<<<<<<<<<<<<<
 *             flags |= PyBUF_WRITABLE
 *         if format:
 */
    }

    /* "buffers.pxd":157
 *         if writable:
 *             flags |= PyBUF_WRITABLE
 *         if format:             # <<<<<<<<<<<<<<
 *             flags |= PyBUF_FORMAT
 *         PyObject_GetBuffer(ob, &view, flags)
 */
    __pyx_t_1 = (__pyx_v_format != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":158
 *             flags |= PyBUF_WRITABLE
 *         if format:
 *             flags |= PyBUF_FORMAT             # <<<<<<<<<<<<<<
 *         PyObject_GetBuffer(ob, &view, flags)
 *         bptr = view.buf
 */
      __pyx_v_flags = (__pyx_v_flags | PyBUF_FORMAT);

      /* "buffers.pxd":157
 *         if writable:
 *             flags |= PyBUF_WRITABLE
 *         if format:             # <<<<<<<<<<<<<<
 *             flags |= PyBUF_FORMAT
 *         PyObject_GetBuffer(ob, &view, flags)
 */
    }

    /* "buffers.pxd":159
 *         if format:
 *             flags |= PyBUF_FORMAT
 *         PyObject_GetBuffer(ob, &view, flags)             # <<<<<<<<<<<<<<
 *         bptr = view.buf
 *         blen = view.len
 */
    __pyx_t_4 = PyObject_GetBuffer(__pyx_v_ob, (&__pyx_v_view), __pyx_v_flags); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 159, __pyx_L1_error)

    /* "buffers.pxd":160
 *             flags |= PyBUF_FORMAT
 *         PyObject_GetBuffer(ob, &view, flags)
 *         bptr = view.buf             # <<<<<<<<<<<<<<
 *         blen = view.len
 *         if format:
 */
    __pyx_t_5 = __pyx_v_view.buf;
    __pyx_v_bptr = __pyx_t_5;

    /* "buffers.pxd":161
 *         PyObject_GetBuffer(ob, &view, flags)
 *         bptr = view.buf
 *         blen = view.len             # <<<<<<<<<<<<<<
 *         if format:
 *             if view.format != NULL:
 */
    __pyx_t_6 = __pyx_v_view.len;
    __pyx_v_blen = __pyx_t_6;

    /* "buffers.pxd":162
 *         bptr = view.buf
 *         blen = view.len
 *         if format:             # <<<<<<<<<<<<<<
 *             if view.format != NULL:
 *                 bfmt = view.format
 */
    __pyx_t_1 = (__pyx_v_format != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":163
 *         blen = view.len
 *         if format:
 *             if view.format != NULL:             # <<<<<<<<<<<<<<
 *                 bfmt = view.format
 *                 bitemlen = view.itemsize
 */
      __pyx_t_1 = ((__pyx_v_view.format != NULL) != 0);
      if (__pyx_t_1) {

        /* "buffers.pxd":164
 *         if format:
 *             if view.format != NULL:
 *                 bfmt = view.format             # <<<<<<<<<<<<<<
 *                 bitemlen = view.itemsize
 *         PyBuffer_Release(&view)
 */
        __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_view.format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_bfmt, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "buffers.pxd":165
 *             if view.format != NULL:
 *                 bfmt = view.format
 *                 bitemlen = view.itemsize             # <<<<<<<<<<<<<<
 *         PyBuffer_Release(&view)
 *     else: # oldstyle
 */
        __pyx_t_6 = __pyx_v_view.itemsize;
        __pyx_v_bitemlen = __pyx_t_6;

        /* "buffers.pxd":163
 *         blen = view.len
 *         if format:
 *             if view.format != NULL:             # <<<<<<<<<<<<<<
 *                 bfmt = view.format
 *                 bitemlen = view.itemsize
 */
      }

      /* "buffers.pxd":162
 *         bptr = view.buf
 *         blen = view.len
 *         if format:             # <<<<<<<<<<<<<<
 *             if view.format != NULL:
 *                 bfmt = view.format
 */
    }

    /* "buffers.pxd":166
 *                 bfmt = view.format
 *                 bitemlen = view.itemsize
 *         PyBuffer_Release(&view)             # <<<<<<<<<<<<<<
 *     else: # oldstyle
 *         if writable:
 */
    PyBuffer_Release((&__pyx_v_view));

    /* "buffers.pxd":153
 *         raise TypeError("%r does not provide a buffer interface."%ob)
 * 
 *     if mode == 3:             # <<<<<<<<<<<<<<
 *         flags = PyBUF_ANY_CONTIGUOUS
 *         if writable:
 */
    goto __pyx_L4;
  }

  /* "buffers.pxd":168
 *         PyBuffer_Release(&view)
 *     else: # oldstyle
 *         if writable:             # <<<<<<<<<<<<<<
 *             PyObject_AsWriteBuffer(ob, &bptr, &blen)
 *         else:
 */
  /*else*/ {
    __pyx_t_1 = (__pyx_v_writable != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":169
 *     else: # oldstyle
 *         if writable:
 *             PyObject_AsWriteBuffer(ob, &bptr, &blen)             # <<<<<<<<<<<<<<
 *         else:
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 */
      __pyx_t_4 = PyObject_AsWriteBuffer(__pyx_v_ob, (&__pyx_v_bptr), (&__pyx_v_blen)); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 169, __pyx_L1_error)

      /* "buffers.pxd":168
 *         PyBuffer_Release(&view)
 *     else: # oldstyle
 *         if writable:             # <<<<<<<<<<<<<<
 *             PyObject_AsWriteBuffer(ob, &bptr, &blen)
 *         else:
 */
      goto __pyx_L9;
    }

    /* "buffers.pxd":171
 *             PyObject_AsWriteBuffer(ob, &bptr, &blen)
 *         else:
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)             # <<<<<<<<<<<<<<
 *         if format:
 *             try: # numpy.ndarray
 */
    /*else*/ {
      __pyx_t_4 = PyObject_AsReadBuffer(__pyx_v_ob, ((const void **)(&__pyx_v_bptr)), (&__pyx_v_blen)); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 171, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "buffers.pxd":172
 *         else:
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 *         if format:             # <<<<<<<<<<<<<<
 *             try: # numpy.ndarray
 *                 dtype = ob.dtype
 */
    __pyx_t_1 = (__pyx_v_format != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":173
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 *         if format:
 *             try: # numpy.ndarray             # <<<<<<<<<<<<<<
 *                 dtype = ob.dtype
 *                 bfmt = dtype.char
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "buffers.pxd":174
 *         if format:
 *             try: # numpy.ndarray
 *                 dtype = ob.dtype             # <<<<<<<<<<<<<<
 *                 bfmt = dtype.char
 *                 bitemlen = dtype.itemsize
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ob, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 174, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_v_dtype = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "buffers.pxd":175
 *             try: # numpy.ndarray
 *                 dtype = ob.dtype
 *                 bfmt = dtype.char             # <<<<<<<<<<<<<<
 *                 bitemlen = dtype.itemsize
 *             except AttributeError:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_char); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 175, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_bfmt, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "buffers.pxd":176
 *                 dtype = ob.dtype
 *                 bfmt = dtype.char
 *                 bitemlen = dtype.itemsize             # <<<<<<<<<<<<<<
 *             except AttributeError:
 *                 try: # array.array
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 176, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_bitemlen = __pyx_t_6;

          /* "buffers.pxd":173
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 *         if format:
 *             try: # numpy.ndarray             # <<<<<<<<<<<<<<
 *                 dtype = ob.dtype
 *                 bfmt = dtype.char
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L18_try_end;
        __pyx_L11_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "buffers.pxd":177
 *                 bfmt = dtype.char
 *                 bitemlen = dtype.itemsize
 *             except AttributeError:             # <<<<<<<<<<<<<<
 *                 try: # array.array
 *                     bfmt = ob.typecode
 */
        __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
        if (__pyx_t_4) {
          __Pyx_AddTraceback("buffers.asbuffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_10) < 0) __PYX_ERR(1, 177, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_10);

          /* "buffers.pxd":178
 *                 bitemlen = dtype.itemsize
 *             except AttributeError:
 *                 try: # array.array             # <<<<<<<<<<<<<<
 *                     bfmt = ob.typecode
 *                     bitemlen = ob.itemsize
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            /*try:*/ {

              /* "buffers.pxd":179
 *             except AttributeError:
 *                 try: # array.array
 *                     bfmt = ob.typecode             # <<<<<<<<<<<<<<
 *                     bitemlen = ob.itemsize
 *                 except AttributeError:
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_ob, __pyx_n_s_typecode); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 179, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF_SET(__pyx_v_bfmt, __pyx_t_14);
              __pyx_t_14 = 0;

              /* "buffers.pxd":180
 *                 try: # array.array
 *                     bfmt = ob.typecode
 *                     bitemlen = ob.itemsize             # <<<<<<<<<<<<<<
 *                 except AttributeError:
 *                     if isinstance(ob, bytes):
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_ob, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 180, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_14); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 180, __pyx_L21_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_v_bitemlen = __pyx_t_6;

              /* "buffers.pxd":178
 *                 bitemlen = dtype.itemsize
 *             except AttributeError:
 *                 try: # array.array             # <<<<<<<<<<<<<<
 *                     bfmt = ob.typecode
 *                     bitemlen = ob.itemsize
 */
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L28_try_end;
            __pyx_L21_error:;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "buffers.pxd":181
 *                     bfmt = ob.typecode
 *                     bitemlen = ob.itemsize
 *                 except AttributeError:             # <<<<<<<<<<<<<<
 *                     if isinstance(ob, bytes):
 *                         bfmt = b"B"
 */
            __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
            if (__pyx_t_4) {
              __Pyx_AddTraceback("buffers.asbuffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0) __PYX_ERR(1, 181, __pyx_L23_except_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_GOTREF(__pyx_t_16);

              /* "buffers.pxd":182
 *                     bitemlen = ob.itemsize
 *                 except AttributeError:
 *                     if isinstance(ob, bytes):             # <<<<<<<<<<<<<<
 *                         bfmt = b"B"
 *                         bitemlen = 1
 */
              __pyx_t_1 = PyBytes_Check(__pyx_v_ob); 
              __pyx_t_17 = (__pyx_t_1 != 0);
              if (__pyx_t_17) {

                /* "buffers.pxd":183
 *                 except AttributeError:
 *                     if isinstance(ob, bytes):
 *                         bfmt = b"B"             # <<<<<<<<<<<<<<
 *                         bitemlen = 1
 *                     else:
 */
                __Pyx_INCREF(__pyx_n_b_B);
                __Pyx_DECREF_SET(__pyx_v_bfmt, __pyx_n_b_B);

                /* "buffers.pxd":184
 *                     if isinstance(ob, bytes):
 *                         bfmt = b"B"
 *                         bitemlen = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # nothing found
 */
                __pyx_v_bitemlen = 1;

                /* "buffers.pxd":182
 *                     bitemlen = ob.itemsize
 *                 except AttributeError:
 *                     if isinstance(ob, bytes):             # <<<<<<<<<<<<<<
 *                         bfmt = b"B"
 *                         bitemlen = 1
 */
                goto __pyx_L31;
              }

              /* "buffers.pxd":187
 *                     else:
 *                         # nothing found
 *                         bfmt = None             # <<<<<<<<<<<<<<
 *                         bitemlen = 0
 *     if base: base[0] = <void *>bptr
 */
              /*else*/ {
                __Pyx_INCREF(Py_None);
                __Pyx_DECREF_SET(__pyx_v_bfmt, Py_None);

                /* "buffers.pxd":188
 *                         # nothing found
 *                         bfmt = None
 *                         bitemlen = 0             # <<<<<<<<<<<<<<
 *     if base: base[0] = <void *>bptr
 *     if size: size[0] = <Py_ssize_t>blen
 */
                __pyx_v_bitemlen = 0;
              }
              __pyx_L31:;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              goto __pyx_L22_exception_handled;
            }
            goto __pyx_L23_except_error;
            __pyx_L23_except_error:;

            /* "buffers.pxd":178
 *                 bitemlen = dtype.itemsize
 *             except AttributeError:
 *                 try: # array.array             # <<<<<<<<<<<<<<
 *                     bfmt = ob.typecode
 *                     bitemlen = ob.itemsize
 */
            __Pyx_PyThreadState_assign
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            goto __pyx_L13_except_error;
            __pyx_L22_exception_handled:;
            __Pyx_PyThreadState_assign
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            __pyx_L28_try_end:;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L12_exception_handled;
        }
        goto __pyx_L13_except_error;
        __pyx_L13_except_error:;

        /* "buffers.pxd":173
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 *         if format:
 *             try: # numpy.ndarray             # <<<<<<<<<<<<<<
 *                 dtype = ob.dtype
 *                 bfmt = dtype.char
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L12_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L18_try_end:;
      }

      /* "buffers.pxd":172
 *         else:
 *             PyObject_AsReadBuffer(ob, <const_void **>&bptr, &blen)
 *         if format:             # <<<<<<<<<<<<<<
 *             try: # numpy.ndarray
 *                 dtype = ob.dtype
 */
    }
  }
  __pyx_L4:;

  /* "buffers.pxd":189
 *                         bfmt = None
 *                         bitemlen = 0
 *     if base: base[0] = <void *>bptr             # <<<<<<<<<<<<<<
 *     if size: size[0] = <Py_ssize_t>blen
 *     if itemsize: itemsize[0] = <Py_ssize_t>bitemlen
 */
  __pyx_t_17 = (__pyx_v_base != 0);
  if (__pyx_t_17) {
    (__pyx_v_base[0]) = ((void *)__pyx_v_bptr);
  }

  /* "buffers.pxd":190
 *                         bitemlen = 0
 *     if base: base[0] = <void *>bptr
 *     if size: size[0] = <Py_ssize_t>blen             # <<<<<<<<<<<<<<
 *     if itemsize: itemsize[0] = <Py_ssize_t>bitemlen
 * 
 */
  __pyx_t_17 = (__pyx_v_size != 0);
  if (__pyx_t_17) {
    (__pyx_v_size[0]) = ((Py_ssize_t)__pyx_v_blen);
  }

  /* "buffers.pxd":191
 *     if base: base[0] = <void *>bptr
 *     if size: size[0] = <Py_ssize_t>blen
 *     if itemsize: itemsize[0] = <Py_ssize_t>bitemlen             # <<<<<<<<<<<<<<
 * 
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:
 */
  __pyx_t_17 = (__pyx_v_itemsize != 0);
  if (__pyx_t_17) {
    (__pyx_v_itemsize[0]) = ((Py_ssize_t)__pyx_v_bitemlen);
  }

  /* "buffers.pxd":193
 *     if itemsize: itemsize[0] = <Py_ssize_t>bitemlen
 * 
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:             # <<<<<<<<<<<<<<
 *         return bfmt.decode('ascii')
 *     return bfmt
 */
  __pyx_t_1 = ((PY_MAJOR_VERSION >= 3) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_17 = __pyx_t_1;
    goto __pyx_L36_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_v_bfmt != Py_None);
  __pyx_t_18 = (__pyx_t_1 != 0);
  __pyx_t_17 = __pyx_t_18;
  __pyx_L36_bool_binop_done:;
  if (__pyx_t_17) {

    /* "buffers.pxd":194
 * 
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:
 *         return bfmt.decode('ascii')             # <<<<<<<<<<<<<<
 *     return bfmt
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_bfmt, __pyx_n_s_decode); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "buffers.pxd":193
 *     if itemsize: itemsize[0] = <Py_ssize_t>bitemlen
 * 
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:             # <<<<<<<<<<<<<<
 *         return bfmt.decode('ascii')
 *     return bfmt
 */
  }

  /* "buffers.pxd":195
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:
 *         return bfmt.decode('ascii')
 *     return bfmt             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_bfmt);
  __pyx_r = __pyx_v_bfmt;
  goto __pyx_L0;

  /* "buffers.pxd":114
 * 
 * 
 * cdef inline object asbuffer(object ob, int writable, int format,             # <<<<<<<<<<<<<<
 *                             void **base, Py_ssize_t *size,
 *                             Py_ssize_t *itemsize):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("buffers.asbuffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bfmt);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":198
 * 
 * 
 * cdef inline object asbuffer_r(object ob, void **base, Py_ssize_t *size):             # <<<<<<<<<<<<<<
 *     """Wrapper for standard calls to asbuffer with a readonly buffer."""
 *     asbuffer(ob, 0, 0, base, size, NULL)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_asbuffer_r(PyObject *__pyx_v_ob, void **__pyx_v_base, Py_ssize_t *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("asbuffer_r", 0);

  /* "buffers.pxd":200
 * cdef inline object asbuffer_r(object ob, void **base, Py_ssize_t *size):
 *     """Wrapper for standard calls to asbuffer with a readonly buffer."""
 *     asbuffer(ob, 0, 0, base, size, NULL)             # <<<<<<<<<<<<<<
 *     return ob
 * 
 */
  __pyx_t_1 = __pyx_f_7buffers_asbuffer(__pyx_v_ob, 0, 0, __pyx_v_base, __pyx_v_size, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "buffers.pxd":201
 *     """Wrapper for standard calls to asbuffer with a readonly buffer."""
 *     asbuffer(ob, 0, 0, base, size, NULL)
 *     return ob             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ob);
  __pyx_r = __pyx_v_ob;
  goto __pyx_L0;

  /* "buffers.pxd":198
 * 
 * 
 * cdef inline object asbuffer_r(object ob, void **base, Py_ssize_t *size):             # <<<<<<<<<<<<<<
 *     """Wrapper for standard calls to asbuffer with a readonly buffer."""
 *     asbuffer(ob, 0, 0, base, size, NULL)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.asbuffer_r", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":204
 * 
 * 
 * cdef inline object asbuffer_w(object ob, void **base, Py_ssize_t *size):             # <<<<<<<<<<<<<<
 *     """Wrapper for standard calls to asbuffer with a writable buffer."""
 *     asbuffer(ob, 1, 0, base, size, NULL)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_asbuffer_w(PyObject *__pyx_v_ob, void **__pyx_v_base, Py_ssize_t *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("asbuffer_w", 0);

  /* "buffers.pxd":206
 * cdef inline object asbuffer_w(object ob, void **base, Py_ssize_t *size):
 *     """Wrapper for standard calls to asbuffer with a writable buffer."""
 *     asbuffer(ob, 1, 0, base, size, NULL)             # <<<<<<<<<<<<<<
 *     return ob
 * 
 */
  __pyx_t_1 = __pyx_f_7buffers_asbuffer(__pyx_v_ob, 1, 0, __pyx_v_base, __pyx_v_size, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "buffers.pxd":207
 *     """Wrapper for standard calls to asbuffer with a writable buffer."""
 *     asbuffer(ob, 1, 0, base, size, NULL)
 *     return ob             # <<<<<<<<<<<<<<
 * 
 * #------------------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ob);
  __pyx_r = __pyx_v_ob;
  goto __pyx_L0;

  /* "buffers.pxd":204
 * 
 * 
 * cdef inline object asbuffer_w(object ob, void **base, Py_ssize_t *size):             # <<<<<<<<<<<<<<
 *     """Wrapper for standard calls to asbuffer with a writable buffer."""
 *     asbuffer(ob, 1, 0, base, size, NULL)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.asbuffer_w", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":214
 * 
 * 
 * cdef inline object frombuffer_3(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Python 3 version of frombuffer.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_3(void *__pyx_v_ptr, Py_ssize_t __pyx_v_s, int __pyx_v_readonly) {
  Py_buffer __pyx_v_pybuf;
  Py_ssize_t *__pyx_v_shape;
  PyObject *__pyx_v_astr = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1[1];
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("frombuffer_3", 0);

  /* "buffers.pxd":221
 *     """
 *     cdef Py_buffer pybuf
 *     cdef Py_ssize_t *shape = [s]             # <<<<<<<<<<<<<<
 *     cdef str astr=""
 *     PyBuffer_FillInfo(&pybuf, astr, ptr, s, readonly, PyBUF_SIMPLE)
 */
  __pyx_t_1[0] = __pyx_v_s;
  __pyx_v_shape = __pyx_t_1;

  /* "buffers.pxd":222
 *     cdef Py_buffer pybuf
 *     cdef Py_ssize_t *shape = [s]
 *     cdef str astr=""             # <<<<<<<<<<<<<<
 *     PyBuffer_FillInfo(&pybuf, astr, ptr, s, readonly, PyBUF_SIMPLE)
 *     pybuf.format = "B"
 */
  __Pyx_INCREF(__pyx_kp_s__2);
  __pyx_v_astr = __pyx_kp_s__2;

  /* "buffers.pxd":223
 *     cdef Py_ssize_t *shape = [s]
 *     cdef str astr=""
 *     PyBuffer_FillInfo(&pybuf, astr, ptr, s, readonly, PyBUF_SIMPLE)             # <<<<<<<<<<<<<<
 *     pybuf.format = "B"
 *     pybuf.shape = shape
 */
  __pyx_t_2 = PyBuffer_FillInfo((&__pyx_v_pybuf), __pyx_v_astr, __pyx_v_ptr, __pyx_v_s, __pyx_v_readonly, PyBUF_SIMPLE); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(1, 223, __pyx_L1_error)

  /* "buffers.pxd":224
 *     cdef str astr=""
 *     PyBuffer_FillInfo(&pybuf, astr, ptr, s, readonly, PyBUF_SIMPLE)
 *     pybuf.format = "B"             # <<<<<<<<<<<<<<
 *     pybuf.shape = shape
 *     pybuf.ndim = 1
 */
  __pyx_v_pybuf.format = ((char *)"B");

  /* "buffers.pxd":225
 *     PyBuffer_FillInfo(&pybuf, astr, ptr, s, readonly, PyBUF_SIMPLE)
 *     pybuf.format = "B"
 *     pybuf.shape = shape             # <<<<<<<<<<<<<<
 *     pybuf.ndim = 1
 *     return PyMemoryView_FromBuffer(&pybuf)
 */
  __pyx_v_pybuf.shape = __pyx_v_shape;

  /* "buffers.pxd":226
 *     pybuf.format = "B"
 *     pybuf.shape = shape
 *     pybuf.ndim = 1             # <<<<<<<<<<<<<<
 *     return PyMemoryView_FromBuffer(&pybuf)
 * 
 */
  __pyx_v_pybuf.ndim = 1;

  /* "buffers.pxd":227
 *     pybuf.shape = shape
 *     pybuf.ndim = 1
 *     return PyMemoryView_FromBuffer(&pybuf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyMemoryView_FromBuffer((&__pyx_v_pybuf)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "buffers.pxd":214
 * 
 * 
 * cdef inline object frombuffer_3(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Python 3 version of frombuffer.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("buffers.frombuffer_3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_astr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":230
 * 
 * 
 * cdef inline object frombuffer_2(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Python 2 version of frombuffer.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_2(void *__pyx_v_ptr, Py_ssize_t __pyx_v_s, int __pyx_v_readonly) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("frombuffer_2", 0);

  /* "buffers.pxd":236
 *     """
 * 
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         if readonly:
 *             return PyBuffer_FromMemory(ptr, s)
 */
  __pyx_t_1 = (__pyx_f_7buffers_oldstyle_available() != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":237
 * 
 *     if oldstyle_available():
 *         if readonly:             # <<<<<<<<<<<<<<
 *             return PyBuffer_FromMemory(ptr, s)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_readonly != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":238
 *     if oldstyle_available():
 *         if readonly:
 *             return PyBuffer_FromMemory(ptr, s)             # <<<<<<<<<<<<<<
 *         else:
 *             return PyBuffer_FromReadWriteMemory(ptr, s)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyBuffer_FromMemory(__pyx_v_ptr, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "buffers.pxd":237
 * 
 *     if oldstyle_available():
 *         if readonly:             # <<<<<<<<<<<<<<
 *             return PyBuffer_FromMemory(ptr, s)
 *         else:
 */
    }

    /* "buffers.pxd":240
 *             return PyBuffer_FromMemory(ptr, s)
 *         else:
 *             return PyBuffer_FromReadWriteMemory(ptr, s)             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError("Old style buffers not available.")
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyBuffer_FromReadWriteMemory(__pyx_v_ptr, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "buffers.pxd":236
 *     """
 * 
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         if readonly:
 *             return PyBuffer_FromMemory(ptr, s)
 */
  }

  /* "buffers.pxd":242
 *             return PyBuffer_FromReadWriteMemory(ptr, s)
 *     else:
 *         raise NotImplementedError("Old style buffers not available.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 242, __pyx_L1_error)
  }

  /* "buffers.pxd":230
 * 
 * 
 * cdef inline object frombuffer_2(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Python 2 version of frombuffer.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("buffers.frombuffer_2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":245
 * 
 * 
 * cdef inline object frombuffer(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Create a Python Buffer/View of a C array.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer(void *__pyx_v_ptr, Py_ssize_t __pyx_v_s, int __pyx_v_readonly) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("frombuffer", 0);

  /* "buffers.pxd":262
 *     """
 *     # oldstyle first priority for now
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         return frombuffer_2(ptr, s, readonly)
 *     else:
 */
  __pyx_t_1 = (__pyx_f_7buffers_oldstyle_available() != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":263
 *     # oldstyle first priority for now
 *     if oldstyle_available():
 *         return frombuffer_2(ptr, s, readonly)             # <<<<<<<<<<<<<<
 *     else:
 *         return frombuffer_3(ptr, s, readonly)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_7buffers_frombuffer_2(__pyx_v_ptr, __pyx_v_s, __pyx_v_readonly); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "buffers.pxd":262
 *     """
 *     # oldstyle first priority for now
 *     if oldstyle_available():             # <<<<<<<<<<<<<<
 *         return frombuffer_2(ptr, s, readonly)
 *     else:
 */
  }

  /* "buffers.pxd":265
 *         return frombuffer_2(ptr, s, readonly)
 *     else:
 *         return frombuffer_3(ptr, s, readonly)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_7buffers_frombuffer_3(__pyx_v_ptr, __pyx_v_s, __pyx_v_readonly); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "buffers.pxd":245
 * 
 * 
 * cdef inline object frombuffer(void *ptr, Py_ssize_t s, int readonly):             # <<<<<<<<<<<<<<
 *     """Create a Python Buffer/View of a C array.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("buffers.frombuffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":268
 * 
 * 
 * cdef inline object frombuffer_r(void *ptr, Py_ssize_t s):             # <<<<<<<<<<<<<<
 *     """Wrapper for readonly view frombuffer."""
 *     return frombuffer(ptr, s, 1)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_r(void *__pyx_v_ptr, Py_ssize_t __pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("frombuffer_r", 0);

  /* "buffers.pxd":270
 * cdef inline object frombuffer_r(void *ptr, Py_ssize_t s):
 *     """Wrapper for readonly view frombuffer."""
 *     return frombuffer(ptr, s, 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7buffers_frombuffer(__pyx_v_ptr, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "buffers.pxd":268
 * 
 * 
 * cdef inline object frombuffer_r(void *ptr, Py_ssize_t s):             # <<<<<<<<<<<<<<
 *     """Wrapper for readonly view frombuffer."""
 *     return frombuffer(ptr, s, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.frombuffer_r", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":273
 * 
 * 
 * cdef inline object frombuffer_w(void *ptr, Py_ssize_t s):             # <<<<<<<<<<<<<<
 *     """Wrapper for writable view frombuffer."""
 *     return frombuffer(ptr, s, 0)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_frombuffer_w(void *__pyx_v_ptr, Py_ssize_t __pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("frombuffer_w", 0);

  /* "buffers.pxd":275
 * cdef inline object frombuffer_w(void *ptr, Py_ssize_t s):
 *     """Wrapper for writable view frombuffer."""
 *     return frombuffer(ptr, s, 0)             # <<<<<<<<<<<<<<
 * 
 * #------------------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7buffers_frombuffer(__pyx_v_ptr, __pyx_v_s, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "buffers.pxd":273
 * 
 * 
 * cdef inline object frombuffer_w(void *ptr, Py_ssize_t s):             # <<<<<<<<<<<<<<
 *     """Wrapper for writable view frombuffer."""
 *     return frombuffer(ptr, s, 0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.frombuffer_w", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":282
 * #------------------------------------------------------------------------------
 * 
 * cdef inline object viewfromobject(object obj, int readonly):             # <<<<<<<<<<<<<<
 *     """Construct a Python Buffer/View object from another Python object.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_viewfromobject(PyObject *__pyx_v_obj, int __pyx_v_readonly) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("viewfromobject", 0);

  /* "buffers.pxd":298
 *     Buffer/View of the original object.
 *     """
 *     if not memoryview_available():             # <<<<<<<<<<<<<<
 *         if readonly:
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)
 */
  __pyx_t_1 = ((!(__pyx_f_7buffers_memoryview_available() != 0)) != 0);
  if (__pyx_t_1) {

    /* "buffers.pxd":299
 *     """
 *     if not memoryview_available():
 *         if readonly:             # <<<<<<<<<<<<<<
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_readonly != 0);
    if (__pyx_t_1) {

      /* "buffers.pxd":300
 *     if not memoryview_available():
 *         if readonly:
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)             # <<<<<<<<<<<<<<
 *         else:
 *             return PyBuffer_FromReadWriteObject(obj, 0, Py_END_OF_BUFFER)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyBuffer_FromObject(__pyx_v_obj, 0, Py_END_OF_BUFFER); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "buffers.pxd":299
 *     """
 *     if not memoryview_available():
 *         if readonly:             # <<<<<<<<<<<<<<
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)
 *         else:
 */
    }

    /* "buffers.pxd":302
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)
 *         else:
 *             return PyBuffer_FromReadWriteObject(obj, 0, Py_END_OF_BUFFER)             # <<<<<<<<<<<<<<
 *     else:
 *         return PyMemoryView_FromObject(obj)
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyBuffer_FromReadWriteObject(__pyx_v_obj, 0, Py_END_OF_BUFFER); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "buffers.pxd":298
 *     Buffer/View of the original object.
 *     """
 *     if not memoryview_available():             # <<<<<<<<<<<<<<
 *         if readonly:
 *             return PyBuffer_FromObject(obj, 0, Py_END_OF_BUFFER)
 */
  }

  /* "buffers.pxd":304
 *             return PyBuffer_FromReadWriteObject(obj, 0, Py_END_OF_BUFFER)
 *     else:
 *         return PyMemoryView_FromObject(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyMemoryView_FromObject(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "buffers.pxd":282
 * #------------------------------------------------------------------------------
 * 
 * cdef inline object viewfromobject(object obj, int readonly):             # <<<<<<<<<<<<<<
 *     """Construct a Python Buffer/View object from another Python object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("buffers.viewfromobject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":307
 * 
 * 
 * cdef inline object viewfromobject_r(object obj):             # <<<<<<<<<<<<<<
 *     """Wrapper for readonly viewfromobject."""
 *     return viewfromobject(obj, 1)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_viewfromobject_r(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("viewfromobject_r", 0);

  /* "buffers.pxd":309
 * cdef inline object viewfromobject_r(object obj):
 *     """Wrapper for readonly viewfromobject."""
 *     return viewfromobject(obj, 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7buffers_viewfromobject(__pyx_v_obj, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "buffers.pxd":307
 * 
 * 
 * cdef inline object viewfromobject_r(object obj):             # <<<<<<<<<<<<<<
 *     """Wrapper for readonly viewfromobject."""
 *     return viewfromobject(obj, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.viewfromobject_r", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "buffers.pxd":312
 * 
 * 
 * cdef inline object viewfromobject_w(object obj):             # <<<<<<<<<<<<<<
 *     """Wrapper for writable viewfromobject."""
 *     return viewfromobject(obj, 0)
 */

static CYTHON_INLINE PyObject *__pyx_f_7buffers_viewfromobject_w(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("viewfromobject_w", 0);

  /* "buffers.pxd":314
 * cdef inline object viewfromobject_w(object obj):
 *     """Wrapper for writable viewfromobject."""
 *     return viewfromobject(obj, 0)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7buffers_viewfromobject(__pyx_v_obj, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "buffers.pxd":312
 * 
 * 
 * cdef inline object viewfromobject_w(object obj):             # <<<<<<<<<<<<<<
 *     """Wrapper for writable viewfromobject."""
 *     return viewfromobject(obj, 0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("buffers.viewfromobject_w", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "zmq/devices/monitoredqueue.pxd":39
 * #-----------------------------------------------------------------------------
 * 
 * cdef inline int _relay(void *insocket_, void *outsocket_, void *sidesocket_,             # <<<<<<<<<<<<<<
 *                 zmq_msg_t msg, zmq_msg_t side_msg, zmq_msg_t id_msg,
 *                 bint swap_ids) nogil:
 */

static CYTHON_INLINE int __pyx_f_3zmq_7devices_14monitoredqueue__relay(void *__pyx_v_insocket_, void *__pyx_v_outsocket_, void *__pyx_v_sidesocket_, zmq_msg_t __pyx_v_msg, zmq_msg_t __pyx_v_side_msg, zmq_msg_t __pyx_v_id_msg, int __pyx_v_swap_ids) {
  int __pyx_v_rc;
  pyzmq_int64_t __pyx_v_flag_2;
  int __pyx_v_flag_3;
  int __pyx_v_flags;
  size_t __pyx_v_flagsz;
  void *__pyx_v_flag_ptr;
  int __pyx_r;
  int __pyx_t_1;

  /* "zmq/devices/monitoredqueue.pxd":50
 *     cdef void * flag_ptr
 * 
 *     if ZMQ_VERSION_MAJOR < 3:             # <<<<<<<<<<<<<<
 *         flagsz = sizeof (int64_t)
 *         flag_ptr = &flag_2
 */
  __pyx_t_1 = ((ZMQ_VERSION_MAJOR < 3) != 0);
  if (__pyx_t_1) {

    /* "zmq/devices/monitoredqueue.pxd":51
 * 
 *     if ZMQ_VERSION_MAJOR < 3:
 *         flagsz = sizeof (int64_t)             # <<<<<<<<<<<<<<
 *         flag_ptr = &flag_2
 *     else:
 */
    __pyx_v_flagsz = (sizeof(pyzmq_int64_t));

    /* "zmq/devices/monitoredqueue.pxd":52
 *     if ZMQ_VERSION_MAJOR < 3:
 *         flagsz = sizeof (int64_t)
 *         flag_ptr = &flag_2             # <<<<<<<<<<<<<<
 *     else:
 *         flagsz = sizeof (int)
 */
    __pyx_v_flag_ptr = (&__pyx_v_flag_2);

    /* "zmq/devices/monitoredqueue.pxd":50
 *     cdef void * flag_ptr
 * 
 *     if ZMQ_VERSION_MAJOR < 3:             # <<<<<<<<<<<<<<
 *         flagsz = sizeof (int64_t)
 *         flag_ptr = &flag_2
 */
    goto __pyx_L3;
  }

  /* "zmq/devices/monitoredqueue.pxd":54
 *         flag_ptr = &flag_2
 *     else:
 *         flagsz = sizeof (int)             # <<<<<<<<<<<<<<
 *         flag_ptr = &flag_3
 * 
 */
  /*else*/ {
    __pyx_v_flagsz = (sizeof(int));

    /* "zmq/devices/monitoredqueue.pxd":55
 *     else:
 *         flagsz = sizeof (int)
 *         flag_ptr = &flag_3             # <<<<<<<<<<<<<<
 * 
 *     if swap_ids:# both router, must send second identity first
 */
    __pyx_v_flag_ptr = (&__pyx_v_flag_3);
  }
  __pyx_L3:;

  /* "zmq/devices/monitoredqueue.pxd":57
 *         flag_ptr = &flag_3
 * 
 *     if swap_ids:# both router, must send second identity first             # <<<<<<<<<<<<<<
 *         # recv two ids into msg, id_msg
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 */
  __pyx_t_1 = (__pyx_v_swap_ids != 0);
  if (__pyx_t_1) {

    /* "zmq/devices/monitoredqueue.pxd":59
 *     if swap_ids:# both router, must send second identity first
 *         # recv two ids into msg, id_msg
 *         rc = zmq_msg_recv(&msg, insocket_, 0)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 * 
 */
    __pyx_v_rc = zmq_msg_recv((&__pyx_v_msg), __pyx_v_insocket_, 0);

    /* "zmq/devices/monitoredqueue.pxd":60
 *         # recv two ids into msg, id_msg
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 * 
 *         rc = zmq_msg_recv(&id_msg, insocket_, 0)
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":62
 *         if rc < 0: return rc
 * 
 *         rc = zmq_msg_recv(&id_msg, insocket_, 0)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 * 
 */
    __pyx_v_rc = zmq_msg_recv((&__pyx_v_id_msg), __pyx_v_insocket_, 0);

    /* "zmq/devices/monitoredqueue.pxd":63
 * 
 *         rc = zmq_msg_recv(&id_msg, insocket_, 0)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 * 
 *         # send second id (id_msg) first
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":67
 *         # send second id (id_msg) first
 *         #!!!! always send a copy before the original !!!!
 *         rc = zmq_msg_copy(&side_msg, &id_msg)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 */
    __pyx_v_rc = zmq_msg_copy((&__pyx_v_side_msg), (&__pyx_v_id_msg));

    /* "zmq/devices/monitoredqueue.pxd":68
 *         #!!!! always send a copy before the original !!!!
 *         rc = zmq_msg_copy(&side_msg, &id_msg)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":69
 *         rc = zmq_msg_copy(&side_msg, &id_msg)
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&id_msg, sidesocket_, ZMQ_SNDMORE)
 */
    __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_outsocket_, ZMQ_SNDMORE);

    /* "zmq/devices/monitoredqueue.pxd":70
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         rc = zmq_msg_send(&id_msg, sidesocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":71
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&id_msg, sidesocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         # send first id (msg) second
 */
    __pyx_v_rc = zmq_msg_send((&__pyx_v_id_msg), __pyx_v_sidesocket_, ZMQ_SNDMORE);

    /* "zmq/devices/monitoredqueue.pxd":72
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&id_msg, sidesocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         # send first id (msg) second
 *         rc = zmq_msg_copy(&side_msg, &msg)
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":74
 *         if rc < 0: return rc
 *         # send first id (msg) second
 *         rc = zmq_msg_copy(&side_msg, &msg)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 */
    __pyx_v_rc = zmq_msg_copy((&__pyx_v_side_msg), (&__pyx_v_msg));

    /* "zmq/devices/monitoredqueue.pxd":75
 *         # send first id (msg) second
 *         rc = zmq_msg_copy(&side_msg, &msg)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":76
 *         rc = zmq_msg_copy(&side_msg, &msg)
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 */
    __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_outsocket_, ZMQ_SNDMORE);

    /* "zmq/devices/monitoredqueue.pxd":77
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":78
 *         rc = zmq_msg_send(&side_msg, outsocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *     while (True):
 */
    __pyx_v_rc = zmq_msg_send((&__pyx_v_msg), __pyx_v_sidesocket_, ZMQ_SNDMORE);

    /* "zmq/devices/monitoredqueue.pxd":79
 *         if rc < 0: return rc
 *         rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *     while (True):
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":57
 *         flag_ptr = &flag_3
 * 
 *     if swap_ids:# both router, must send second identity first             # <<<<<<<<<<<<<<
 *         # recv two ids into msg, id_msg
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 */
  }

  /* "zmq/devices/monitoredqueue.pxd":80
 *         rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 *         if rc < 0: return rc
 *     while (True):             # <<<<<<<<<<<<<<
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 *         if rc < 0: return rc
 */
  while (1) {

    /* "zmq/devices/monitoredqueue.pxd":81
 *         if rc < 0: return rc
 *     while (True):
 *         rc = zmq_msg_recv(&msg, insocket_, 0)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         # assert (rc == 0)
 */
    __pyx_v_rc = zmq_msg_recv((&__pyx_v_msg), __pyx_v_insocket_, 0);

    /* "zmq/devices/monitoredqueue.pxd":82
 *     while (True):
 *         rc = zmq_msg_recv(&msg, insocket_, 0)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         # assert (rc == 0)
 *         rc = zmq_getsockopt (insocket_, ZMQ_RCVMORE, flag_ptr, &flagsz)
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":84
 *         if rc < 0: return rc
 *         # assert (rc == 0)
 *         rc = zmq_getsockopt (insocket_, ZMQ_RCVMORE, flag_ptr, &flagsz)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         flags = 0
 */
    __pyx_v_rc = zmq_getsockopt(__pyx_v_insocket_, ZMQ_RCVMORE, __pyx_v_flag_ptr, (&__pyx_v_flagsz));

    /* "zmq/devices/monitoredqueue.pxd":85
 *         # assert (rc == 0)
 *         rc = zmq_getsockopt (insocket_, ZMQ_RCVMORE, flag_ptr, &flagsz)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         flags = 0
 *         if ZMQ_VERSION_MAJOR < 3:
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":86
 *         rc = zmq_getsockopt (insocket_, ZMQ_RCVMORE, flag_ptr, &flagsz)
 *         if rc < 0: return rc
 *         flags = 0             # <<<<<<<<<<<<<<
 *         if ZMQ_VERSION_MAJOR < 3:
 *             if flag_2:
 */
    __pyx_v_flags = 0;

    /* "zmq/devices/monitoredqueue.pxd":87
 *         if rc < 0: return rc
 *         flags = 0
 *         if ZMQ_VERSION_MAJOR < 3:             # <<<<<<<<<<<<<<
 *             if flag_2:
 *                 flags |= ZMQ_SNDMORE
 */
    __pyx_t_1 = ((ZMQ_VERSION_MAJOR < 3) != 0);
    if (__pyx_t_1) {

      /* "zmq/devices/monitoredqueue.pxd":88
 *         flags = 0
 *         if ZMQ_VERSION_MAJOR < 3:
 *             if flag_2:             # <<<<<<<<<<<<<<
 *                 flags |= ZMQ_SNDMORE
 *         else:
 */
      __pyx_t_1 = (__pyx_v_flag_2 != 0);
      if (__pyx_t_1) {

        /* "zmq/devices/monitoredqueue.pxd":89
 *         if ZMQ_VERSION_MAJOR < 3:
 *             if flag_2:
 *                 flags |= ZMQ_SNDMORE             # <<<<<<<<<<<<<<
 *         else:
 *             if flag_3:
 */
        __pyx_v_flags = (__pyx_v_flags | ZMQ_SNDMORE);

        /* "zmq/devices/monitoredqueue.pxd":88
 *         flags = 0
 *         if ZMQ_VERSION_MAJOR < 3:
 *             if flag_2:             # <<<<<<<<<<<<<<
 *                 flags |= ZMQ_SNDMORE
 *         else:
 */
      }

      /* "zmq/devices/monitoredqueue.pxd":87
 *         if rc < 0: return rc
 *         flags = 0
 *         if ZMQ_VERSION_MAJOR < 3:             # <<<<<<<<<<<<<<
 *             if flag_2:
 *                 flags |= ZMQ_SNDMORE
 */
      goto __pyx_L17;
    }

    /* "zmq/devices/monitoredqueue.pxd":91
 *                 flags |= ZMQ_SNDMORE
 *         else:
 *             if flag_3:             # <<<<<<<<<<<<<<
 *                 flags |= ZMQ_SNDMORE
 *             # LABEL has been removed:
 */
    /*else*/ {
      __pyx_t_1 = (__pyx_v_flag_3 != 0);
      if (__pyx_t_1) {

        /* "zmq/devices/monitoredqueue.pxd":92
 *         else:
 *             if flag_3:
 *                 flags |= ZMQ_SNDMORE             # <<<<<<<<<<<<<<
 *             # LABEL has been removed:
 *             # rc = zmq_getsockopt (insocket_, ZMQ_RCVLABEL, flag_ptr, &flagsz)
 */
        __pyx_v_flags = (__pyx_v_flags | ZMQ_SNDMORE);

        /* "zmq/devices/monitoredqueue.pxd":91
 *                 flags |= ZMQ_SNDMORE
 *         else:
 *             if flag_3:             # <<<<<<<<<<<<<<
 *                 flags |= ZMQ_SNDMORE
 *             # LABEL has been removed:
 */
      }
    }
    __pyx_L17:;

    /* "zmq/devices/monitoredqueue.pxd":99
 *         # assert (rc == 0)
 * 
 *         rc = zmq_msg_copy(&side_msg, &msg)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         if flags:
 */
    __pyx_v_rc = zmq_msg_copy((&__pyx_v_side_msg), (&__pyx_v_msg));

    /* "zmq/devices/monitoredqueue.pxd":100
 * 
 *         rc = zmq_msg_copy(&side_msg, &msg)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         if flags:
 *             rc = zmq_msg_send(&side_msg, outsocket_, flags)
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":101
 *         rc = zmq_msg_copy(&side_msg, &msg)
 *         if rc < 0: return rc
 *         if flags:             # <<<<<<<<<<<<<<
 *             rc = zmq_msg_send(&side_msg, outsocket_, flags)
 *             if rc < 0: return rc
 */
    __pyx_t_1 = (__pyx_v_flags != 0);
    if (__pyx_t_1) {

      /* "zmq/devices/monitoredqueue.pxd":102
 *         if rc < 0: return rc
 *         if flags:
 *             rc = zmq_msg_send(&side_msg, outsocket_, flags)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             # only SNDMORE for side-socket
 */
      __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_outsocket_, __pyx_v_flags);

      /* "zmq/devices/monitoredqueue.pxd":103
 *         if flags:
 *             rc = zmq_msg_send(&side_msg, outsocket_, flags)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             # only SNDMORE for side-socket
 *             rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":105
 *             if rc < 0: return rc
 *             # only SNDMORE for side-socket
 *             rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *         else:
 */
      __pyx_v_rc = zmq_msg_send((&__pyx_v_msg), __pyx_v_sidesocket_, ZMQ_SNDMORE);

      /* "zmq/devices/monitoredqueue.pxd":106
 *             # only SNDMORE for side-socket
 *             rc = zmq_msg_send(&msg, sidesocket_, ZMQ_SNDMORE)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         else:
 *             rc = zmq_msg_send(&side_msg, outsocket_, 0)
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":101
 *         rc = zmq_msg_copy(&side_msg, &msg)
 *         if rc < 0: return rc
 *         if flags:             # <<<<<<<<<<<<<<
 *             rc = zmq_msg_send(&side_msg, outsocket_, flags)
 *             if rc < 0: return rc
 */
      goto __pyx_L21;
    }

    /* "zmq/devices/monitoredqueue.pxd":108
 *             if rc < 0: return rc
 *         else:
 *             rc = zmq_msg_send(&side_msg, outsocket_, 0)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&msg, sidesocket_, 0)
 */
    /*else*/ {
      __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_outsocket_, 0);

      /* "zmq/devices/monitoredqueue.pxd":109
 *         else:
 *             rc = zmq_msg_send(&side_msg, outsocket_, 0)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             rc = zmq_msg_send(&msg, sidesocket_, 0)
 *             if rc < 0: return rc
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":110
 *             rc = zmq_msg_send(&side_msg, outsocket_, 0)
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&msg, sidesocket_, 0)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             break
 */
      __pyx_v_rc = zmq_msg_send((&__pyx_v_msg), __pyx_v_sidesocket_, 0);

      /* "zmq/devices/monitoredqueue.pxd":111
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&msg, sidesocket_, 0)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             break
 *     return rc
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":112
 *             rc = zmq_msg_send(&msg, sidesocket_, 0)
 *             if rc < 0: return rc
 *             break             # <<<<<<<<<<<<<<
 *     return rc
 * 
 */
      goto __pyx_L14_break;
    }
    __pyx_L21:;
  }
  __pyx_L14_break:;

  /* "zmq/devices/monitoredqueue.pxd":113
 *             if rc < 0: return rc
 *             break
 *     return rc             # <<<<<<<<<<<<<<
 * 
 * # the MonitoredQueue C function, adapted from zmq::queue.cpp :
 */
  __pyx_r = __pyx_v_rc;
  goto __pyx_L0;

  /* "zmq/devices/monitoredqueue.pxd":39
 * #-----------------------------------------------------------------------------
 * 
 * cdef inline int _relay(void *insocket_, void *outsocket_, void *sidesocket_,             # <<<<<<<<<<<<<<
 *                 zmq_msg_t msg, zmq_msg_t side_msg, zmq_msg_t id_msg,
 *                 bint swap_ids) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "zmq/devices/monitoredqueue.pxd":116
 * 
 * # the MonitoredQueue C function, adapted from zmq::queue.cpp :
 * cdef inline int c_monitored_queue (void *insocket_, void *outsocket_,             # <<<<<<<<<<<<<<
 *                         void *sidesocket_, zmq_msg_t *in_msg_ptr,
 *                         zmq_msg_t *out_msg_ptr, int swap_ids) nogil:
 */

static CYTHON_INLINE int __pyx_f_3zmq_7devices_14monitoredqueue_c_monitored_queue(void *__pyx_v_insocket_, void *__pyx_v_outsocket_, void *__pyx_v_sidesocket_, zmq_msg_t *__pyx_v_in_msg_ptr, zmq_msg_t *__pyx_v_out_msg_ptr, int __pyx_v_swap_ids) {
  zmq_msg_t __pyx_v_msg;
  int __pyx_v_rc;
  zmq_msg_t __pyx_v_id_msg;
  zmq_msg_t __pyx_v_side_msg;
  zmq_pollitem_t __pyx_v_items[2];
  int __pyx_r;
  int __pyx_t_1;

  /* "zmq/devices/monitoredqueue.pxd":125
 * 
 *     cdef zmq_msg_t msg
 *     cdef int rc = zmq_msg_init (&msg)             # <<<<<<<<<<<<<<
 *     cdef zmq_msg_t id_msg
 *     rc = zmq_msg_init (&id_msg)
 */
  __pyx_v_rc = zmq_msg_init((&__pyx_v_msg));

  /* "zmq/devices/monitoredqueue.pxd":127
 *     cdef int rc = zmq_msg_init (&msg)
 *     cdef zmq_msg_t id_msg
 *     rc = zmq_msg_init (&id_msg)             # <<<<<<<<<<<<<<
 *     if rc < 0: return rc
 *     cdef zmq_msg_t side_msg
 */
  __pyx_v_rc = zmq_msg_init((&__pyx_v_id_msg));

  /* "zmq/devices/monitoredqueue.pxd":128
 *     cdef zmq_msg_t id_msg
 *     rc = zmq_msg_init (&id_msg)
 *     if rc < 0: return rc             # <<<<<<<<<<<<<<
 *     cdef zmq_msg_t side_msg
 *     rc = zmq_msg_init (&side_msg)
 */
  __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_rc;
    goto __pyx_L0;
  }

  /* "zmq/devices/monitoredqueue.pxd":130
 *     if rc < 0: return rc
 *     cdef zmq_msg_t side_msg
 *     rc = zmq_msg_init (&side_msg)             # <<<<<<<<<<<<<<
 *     if rc < 0: return rc
 * 
 */
  __pyx_v_rc = zmq_msg_init((&__pyx_v_side_msg));

  /* "zmq/devices/monitoredqueue.pxd":131
 *     cdef zmq_msg_t side_msg
 *     rc = zmq_msg_init (&side_msg)
 *     if rc < 0: return rc             # <<<<<<<<<<<<<<
 * 
 *     cdef zmq_pollitem_t items [2]
 */
  __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_rc;
    goto __pyx_L0;
  }

  /* "zmq/devices/monitoredqueue.pxd":134
 * 
 *     cdef zmq_pollitem_t items [2]
 *     items [0].socket = insocket_             # <<<<<<<<<<<<<<
 *     items [0].fd = 0
 *     items [0].events = ZMQ_POLLIN
 */
  (__pyx_v_items[0]).socket = __pyx_v_insocket_;

  /* "zmq/devices/monitoredqueue.pxd":135
 *     cdef zmq_pollitem_t items [2]
 *     items [0].socket = insocket_
 *     items [0].fd = 0             # <<<<<<<<<<<<<<
 *     items [0].events = ZMQ_POLLIN
 *     items [0].revents = 0
 */
  (__pyx_v_items[0]).fd = 0;

  /* "zmq/devices/monitoredqueue.pxd":136
 *     items [0].socket = insocket_
 *     items [0].fd = 0
 *     items [0].events = ZMQ_POLLIN             # <<<<<<<<<<<<<<
 *     items [0].revents = 0
 *     items [1].socket = outsocket_
 */
  (__pyx_v_items[0]).events = ZMQ_POLLIN;

  /* "zmq/devices/monitoredqueue.pxd":137
 *     items [0].fd = 0
 *     items [0].events = ZMQ_POLLIN
 *     items [0].revents = 0             # <<<<<<<<<<<<<<
 *     items [1].socket = outsocket_
 *     items [1].fd = 0
 */
  (__pyx_v_items[0]).revents = 0;

  /* "zmq/devices/monitoredqueue.pxd":138
 *     items [0].events = ZMQ_POLLIN
 *     items [0].revents = 0
 *     items [1].socket = outsocket_             # <<<<<<<<<<<<<<
 *     items [1].fd = 0
 *     items [1].events = ZMQ_POLLIN
 */
  (__pyx_v_items[1]).socket = __pyx_v_outsocket_;

  /* "zmq/devices/monitoredqueue.pxd":139
 *     items [0].revents = 0
 *     items [1].socket = outsocket_
 *     items [1].fd = 0             # <<<<<<<<<<<<<<
 *     items [1].events = ZMQ_POLLIN
 *     items [1].revents = 0
 */
  (__pyx_v_items[1]).fd = 0;

  /* "zmq/devices/monitoredqueue.pxd":140
 *     items [1].socket = outsocket_
 *     items [1].fd = 0
 *     items [1].events = ZMQ_POLLIN             # <<<<<<<<<<<<<<
 *     items [1].revents = 0
 *     # I don't think sidesocket should be polled?
 */
  (__pyx_v_items[1]).events = ZMQ_POLLIN;

  /* "zmq/devices/monitoredqueue.pxd":141
 *     items [1].fd = 0
 *     items [1].events = ZMQ_POLLIN
 *     items [1].revents = 0             # <<<<<<<<<<<<<<
 *     # I don't think sidesocket should be polled?
 *     # items [2].socket = sidesocket_
 */
  (__pyx_v_items[1]).revents = 0;

  /* "zmq/devices/monitoredqueue.pxd":148
 *     # items [2].revents = 0
 * 
 *     while (True):             # <<<<<<<<<<<<<<
 * 
 *         # //  Wait while there are either requests or replies to process.
 */
  while (1) {

    /* "zmq/devices/monitoredqueue.pxd":151
 * 
 *         # //  Wait while there are either requests or replies to process.
 *         rc = zmq_poll (&items [0], 2, -1)             # <<<<<<<<<<<<<<
 *         if rc < 0: return rc
 *         # //  The algorithm below asumes ratio of request and replies processed
 */
    __pyx_v_rc = zmq_poll((&(__pyx_v_items[0])), 2, -1L);

    /* "zmq/devices/monitoredqueue.pxd":152
 *         # //  Wait while there are either requests or replies to process.
 *         rc = zmq_poll (&items [0], 2, -1)
 *         if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         # //  The algorithm below asumes ratio of request and replies processed
 *         # //  under full load to be 1:1. Although processing requests replies
 */
    __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
    if (__pyx_t_1) {
      __pyx_r = __pyx_v_rc;
      goto __pyx_L0;
    }

    /* "zmq/devices/monitoredqueue.pxd":159
 *         #
 *         # //  Process a request.
 *         if (items [0].revents & ZMQ_POLLIN):             # <<<<<<<<<<<<<<
 *             # send in_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, in_msg_ptr)
 */
    __pyx_t_1 = (((__pyx_v_items[0]).revents & ZMQ_POLLIN) != 0);
    if (__pyx_t_1) {

      /* "zmq/devices/monitoredqueue.pxd":161
 *         if (items [0].revents & ZMQ_POLLIN):
 *             # send in_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, in_msg_ptr)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 */
      __pyx_v_rc = zmq_msg_copy((&__pyx_v_side_msg), __pyx_v_in_msg_ptr);

      /* "zmq/devices/monitoredqueue.pxd":162
 *             # send in_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, in_msg_ptr)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 *             if rc < 0: return rc
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":163
 *             rc = zmq_msg_copy(&side_msg, in_msg_ptr)
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             # relay the rest of the message
 */
      __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_sidesocket_, ZMQ_SNDMORE);

      /* "zmq/devices/monitoredqueue.pxd":164
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             # relay the rest of the message
 *             rc = _relay(insocket_, outsocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":166
 *             if rc < 0: return rc
 *             # relay the rest of the message
 *             rc = _relay(insocket_, outsocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *         if (items [1].revents & ZMQ_POLLIN):
 */
      __pyx_v_rc = __pyx_f_3zmq_7devices_14monitoredqueue__relay(__pyx_v_insocket_, __pyx_v_outsocket_, __pyx_v_sidesocket_, __pyx_v_msg, __pyx_v_side_msg, __pyx_v_id_msg, __pyx_v_swap_ids);

      /* "zmq/devices/monitoredqueue.pxd":167
 *             # relay the rest of the message
 *             rc = _relay(insocket_, outsocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *         if (items [1].revents & ZMQ_POLLIN):
 *             # send out_prefix to side socket
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":159
 *         #
 *         # //  Process a request.
 *         if (items [0].revents & ZMQ_POLLIN):             # <<<<<<<<<<<<<<
 *             # send in_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, in_msg_ptr)
 */
    }

    /* "zmq/devices/monitoredqueue.pxd":168
 *             rc = _relay(insocket_, outsocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 *             if rc < 0: return rc
 *         if (items [1].revents & ZMQ_POLLIN):             # <<<<<<<<<<<<<<
 *             # send out_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, out_msg_ptr)
 */
    __pyx_t_1 = (((__pyx_v_items[1]).revents & ZMQ_POLLIN) != 0);
    if (__pyx_t_1) {

      /* "zmq/devices/monitoredqueue.pxd":170
 *         if (items [1].revents & ZMQ_POLLIN):
 *             # send out_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, out_msg_ptr)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 */
      __pyx_v_rc = zmq_msg_copy((&__pyx_v_side_msg), __pyx_v_out_msg_ptr);

      /* "zmq/devices/monitoredqueue.pxd":171
 *             # send out_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, out_msg_ptr)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 *             if rc < 0: return rc
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":172
 *             rc = zmq_msg_copy(&side_msg, out_msg_ptr)
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *             # relay the rest of the message
 */
      __pyx_v_rc = zmq_msg_send((&__pyx_v_side_msg), __pyx_v_sidesocket_, ZMQ_SNDMORE);

      /* "zmq/devices/monitoredqueue.pxd":173
 *             if rc < 0: return rc
 *             rc = zmq_msg_send(&side_msg, sidesocket_, ZMQ_SNDMORE)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *             # relay the rest of the message
 *             rc = _relay(outsocket_, insocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":175
 *             if rc < 0: return rc
 *             # relay the rest of the message
 *             rc = _relay(outsocket_, insocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)             # <<<<<<<<<<<<<<
 *             if rc < 0: return rc
 *     return rc
 */
      __pyx_v_rc = __pyx_f_3zmq_7devices_14monitoredqueue__relay(__pyx_v_outsocket_, __pyx_v_insocket_, __pyx_v_sidesocket_, __pyx_v_msg, __pyx_v_side_msg, __pyx_v_id_msg, __pyx_v_swap_ids);

      /* "zmq/devices/monitoredqueue.pxd":176
 *             # relay the rest of the message
 *             rc = _relay(outsocket_, insocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 *             if rc < 0: return rc             # <<<<<<<<<<<<<<
 *     return rc
 */
      __pyx_t_1 = ((__pyx_v_rc < 0) != 0);
      if (__pyx_t_1) {
        __pyx_r = __pyx_v_rc;
        goto __pyx_L0;
      }

      /* "zmq/devices/monitoredqueue.pxd":168
 *             rc = _relay(insocket_, outsocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 *             if rc < 0: return rc
 *         if (items [1].revents & ZMQ_POLLIN):             # <<<<<<<<<<<<<<
 *             # send out_prefix to side socket
 *             rc = zmq_msg_copy(&side_msg, out_msg_ptr)
 */
    }
  }

  /* "zmq/devices/monitoredqueue.pxd":177
 *             rc = _relay(outsocket_, insocket_, sidesocket_, msg, side_msg, id_msg, swap_ids)
 *             if rc < 0: return rc
 *     return rc             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_rc;
  goto __pyx_L0;

  /* "zmq/devices/monitoredqueue.pxd":116
 * 
 * # the MonitoredQueue C function, adapted from zmq::queue.cpp :
 * cdef inline int c_monitored_queue (void *insocket_, void *outsocket_,             # <<<<<<<<<<<<<<
 *                         void *sidesocket_, zmq_msg_t *in_msg_ptr,
 *                         zmq_msg_t *out_msg_ptr, int swap_ids) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "zmq/backend/cython/checkrc.pxd":6
 * 
 * 
 * cdef inline int _check_rc(int rc) except -1:             # <<<<<<<<<<<<<<
 *     """internal utility for checking zmq return condition
 * 
 */

static CYTHON_INLINE int __pyx_f_3zmq_7backend_6cython_7checkrc__check_rc(int __pyx_v_rc) {
  int __pyx_v_errno;
  PyObject *__pyx_v_InterruptedSystemCall = NULL;
  PyObject *__pyx_v_Again = NULL;
  PyObject *__pyx_v_ContextTerminated = NULL;
  PyObject *__pyx_v_ZMQError = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_check_rc", 0);

  /* "zmq/backend/cython/checkrc.pxd":11
 *     and raising the appropriate Exception class
 *     """
 *     cdef int errno = zmq_errno()             # <<<<<<<<<<<<<<
 *     PyErr_CheckSignals()
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?
 */
  __pyx_v_errno = zmq_errno();

  /* "zmq/backend/cython/checkrc.pxd":12
 *     """
 *     cdef int errno = zmq_errno()
 *     PyErr_CheckSignals()             # <<<<<<<<<<<<<<
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?
 *         if errno == EINTR:
 */
  __pyx_t_1 = PyErr_CheckSignals(); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(2, 12, __pyx_L1_error)

  /* "zmq/backend/cython/checkrc.pxd":13
 *     cdef int errno = zmq_errno()
 *     PyErr_CheckSignals()
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?             # <<<<<<<<<<<<<<
 *         if errno == EINTR:
 *             from zmq.error import InterruptedSystemCall
 */
  __pyx_t_2 = ((__pyx_v_rc == -1L) != 0);
  if (__pyx_t_2) {

    /* "zmq/backend/cython/checkrc.pxd":14
 *     PyErr_CheckSignals()
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?
 *         if errno == EINTR:             # <<<<<<<<<<<<<<
 *             from zmq.error import InterruptedSystemCall
 *             raise InterruptedSystemCall(errno)
 */
    switch (__pyx_v_errno) {
      case EINTR:

      /* "zmq/backend/cython/checkrc.pxd":15
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?
 *         if errno == EINTR:
 *             from zmq.error import InterruptedSystemCall             # <<<<<<<<<<<<<<
 *             raise InterruptedSystemCall(errno)
 *         elif errno == EAGAIN:
 */
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_n_s_InterruptedSystemCall);
      __Pyx_GIVEREF(__pyx_n_s_InterruptedSystemCall);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_InterruptedSystemCall);
      __pyx_t_4 = __Pyx_Import(__pyx_n_s_zmq_error, __pyx_t_3, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_InterruptedSystemCall); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_InterruptedSystemCall = __pyx_t_3;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "zmq/backend/cython/checkrc.pxd":16
 *         if errno == EINTR:
 *             from zmq.error import InterruptedSystemCall
 *             raise InterruptedSystemCall(errno)             # <<<<<<<<<<<<<<
 *         elif errno == EAGAIN:
 *             from zmq.error import Again
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_errno); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 16, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_InterruptedSystemCall);
      __pyx_t_5 = __pyx_v_InterruptedSystemCall; __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 16, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(2, 16, __pyx_L1_error)

      /* "zmq/backend/cython/checkrc.pxd":14
 *     PyErr_CheckSignals()
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?
 *         if errno == EINTR:             # <<<<<<<<<<<<<<
 *             from zmq.error import InterruptedSystemCall
 *             raise InterruptedSystemCall(errno)
 */
      break;

      /* "zmq/backend/cython/checkrc.pxd":17
 *             from zmq.error import InterruptedSystemCall
 *             raise InterruptedSystemCall(errno)
 *         elif errno == EAGAIN:             # <<<<<<<<<<<<<<
 *             from zmq.error import Again
 *             raise Again(errno)
 */
      case EAGAIN:

      /* "zmq/backend/cython/checkrc.pxd":18
 *             raise InterruptedSystemCall(errno)
 *         elif errno == EAGAIN:
 *             from zmq.error import Again             # <<<<<<<<<<<<<<
 *             raise Again(errno)
 *         elif errno == ZMQ_ETERM:
 */
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 18, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s_Again);
      __Pyx_GIVEREF(__pyx_n_s_Again);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_Again);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_zmq_error, __pyx_t_4, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 18, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_Again); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 18, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_Again = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "zmq/backend/cython/checkrc.pxd":19
 *         elif errno == EAGAIN:
 *             from zmq.error import Again
 *             raise Again(errno)             # <<<<<<<<<<<<<<
 *         elif errno == ZMQ_ETERM:
 *             from zmq.error import ContextTerminated
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_errno); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_Again);
      __pyx_t_7 = __pyx_v_Again; __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 19, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 19, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 19, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(2, 19, __pyx_L1_error)

      /* "zmq/backend/cython/checkrc.pxd":17
 *             from zmq.error import InterruptedSystemCall
 *             raise InterruptedSystemCall(errno)
 *         elif errno == EAGAIN:             # <<<<<<<<<<<<<<
 *             from zmq.error import Again
 *             raise Again(errno)
 */
      break;

      /* "zmq/backend/cython/checkrc.pxd":20
 *             from zmq.error import Again
 *             raise Again(errno)
 *         elif errno == ZMQ_ETERM:             # <<<<<<<<<<<<<<
 *             from zmq.error import ContextTerminated
 *             raise ContextTerminated(errno)
 */
      case ETERM:

      /* "zmq/backend/cython/checkrc.pxd":21
 *             raise Again(errno)
 *         elif errno == ZMQ_ETERM:
 *             from zmq.error import ContextTerminated             # <<<<<<<<<<<<<<
 *             raise ContextTerminated(errno)
 *         else:
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 21, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_s_ContextTerminated);
      __Pyx_GIVEREF(__pyx_n_s_ContextTerminated);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_ContextTerminated);
      __pyx_t_7 = __Pyx_Import(__pyx_n_s_zmq_error, __pyx_t_5, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 21, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_ContextTerminated); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 21, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ContextTerminated = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "zmq/backend/cython/checkrc.pxd":22
 *         elif errno == ZMQ_ETERM:
 *             from zmq.error import ContextTerminated
 *             raise ContextTerminated(errno)             # <<<<<<<<<<<<<<
 *         else:
 *             from zmq.error import ZMQError
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 22, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_ContextTerminated);
      __pyx_t_6 = __pyx_v_ContextTerminated; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 22, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(2, 22, __pyx_L1_error)

      /* "zmq/backend/cython/checkrc.pxd":20
 *             from zmq.error import Again
 *             raise Again(errno)
 *         elif errno == ZMQ_ETERM:             # <<<<<<<<<<<<<<
 *             from zmq.error import ContextTerminated
 *             raise ContextTerminated(errno)
 */
      break;
      default:

      /* "zmq/backend/cython/checkrc.pxd":24
 *             raise ContextTerminated(errno)
 *         else:
 *             from zmq.error import ZMQError             # <<<<<<<<<<<<<<
 *             raise ZMQError(errno)
 *     return 0
 */
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_n_s_ZMQError);
      __Pyx_GIVEREF(__pyx_n_s_ZMQError);
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_ZMQError);
      __pyx_t_6 = __Pyx_Import(__pyx_n_s_zmq_error, __pyx_t_7, -1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_ZMQError); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_ZMQError = __pyx_t_7;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "zmq/backend/cython/checkrc.pxd":25
 *         else:
 *             from zmq.error import ZMQError
 *             raise ZMQError(errno)             # <<<<<<<<<<<<<<
 *     return 0
 */
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_errno); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 25, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_ZMQError);
      __pyx_t_3 = __pyx_v_ZMQError; __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 25, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else {
        __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(2, 25, __pyx_L1_error)
      break;
    }

    /* "zmq/backend/cython/checkrc.pxd":13
 *     cdef int errno = zmq_errno()
 *     PyErr_CheckSignals()
 *     if rc == -1: # if rc < -1, it's a bug in libzmq. Should we warn?             # <<<<<<<<<<<<<<
 *         if errno == EINTR:
 *             from zmq.error import InterruptedSystemCall
 */
  }

  /* "zmq/backend/cython/checkrc.pxd":26
 *             from zmq.error import ZMQError
 *             raise ZMQError(errno)
 *     return 0             # <<<<<<<<<<<<<<
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "zmq/backend/cython/checkrc.pxd":6
 * 
 * 
 * cdef inline int _check_rc(int rc) except -1:             # <<<<<<<<<<<<<<
 *     """internal utility for checking zmq return condition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("zmq.backend.cython.checkrc._check_rc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_InterruptedSystemCall);
  __Pyx_XDECREF(__pyx_v_Again);
  __Pyx_XDECREF(__pyx_v_ContextTerminated);
  __Pyx_XDECREF(__pyx_v_ZMQError);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "monitoredqueue",
    __pyx_k_MonitoredQueue_classes_and_funct, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Again, __pyx_k_Again, sizeof(__pyx_k_Again), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_b_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 0, 1},
  {&__pyx_n_s_ContextTerminated, __pyx_k_ContextTerminated, sizeof(__pyx_k_ContextTerminated), 0, 0, 1, 1},
  {&__pyx_n_s_InterruptedSystemCall, __pyx_k_InterruptedSystemCall, sizeof(__pyx_k_InterruptedSystemCall), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_kp_s_Old_style_buffers_not_available, __pyx_k_Old_style_buffers_not_available, sizeof(__pyx_k_Old_style_buffers_not_available), 0, 0, 1, 0},
  {&__pyx_n_s_ROUTER, __pyx_k_ROUTER, sizeof(__pyx_k_ROUTER), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ZMQError, __pyx_k_ZMQError, sizeof(__pyx_k_ZMQError), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_char, __pyx_k_char, sizeof(__pyx_k_char), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_b_in, __pyx_k_in, sizeof(__pyx_k_in), 0, 0, 0, 1},
  {&__pyx_n_s_in_msg, __pyx_k_in_msg, sizeof(__pyx_k_in_msg), 0, 0, 1, 1},
  {&__pyx_n_s_in_prefix, __pyx_k_in_prefix, sizeof(__pyx_k_in_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_in_socket, __pyx_k_in_socket, sizeof(__pyx_k_in_socket), 0, 0, 1, 1},
  {&__pyx_n_s_ins, __pyx_k_ins, sizeof(__pyx_k_ins), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mon_socket, __pyx_k_mon_socket, sizeof(__pyx_k_mon_socket), 0, 0, 1, 1},
  {&__pyx_n_s_monitored_queue, __pyx_k_monitored_queue, sizeof(__pyx_k_monitored_queue), 0, 0, 1, 1},
  {&__pyx_n_s_mons, __pyx_k_mons, sizeof(__pyx_k_mons), 0, 0, 1, 1},
  {&__pyx_n_s_msg_c, __pyx_k_msg_c, sizeof(__pyx_k_msg_c), 0, 0, 1, 1},
  {&__pyx_n_s_msg_c_len, __pyx_k_msg_c_len, sizeof(__pyx_k_msg_c_len), 0, 0, 1, 1},
  {&__pyx_n_b_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 0, 1},
  {&__pyx_n_s_out_msg, __pyx_k_out_msg, sizeof(__pyx_k_out_msg), 0, 0, 1, 1},
  {&__pyx_n_s_out_prefix, __pyx_k_out_prefix, sizeof(__pyx_k_out_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_out_socket, __pyx_k_out_socket, sizeof(__pyx_k_out_socket), 0, 0, 1, 1},
  {&__pyx_n_s_outs, __pyx_k_outs, sizeof(__pyx_k_outs), 0, 0, 1, 1},
  {&__pyx_kp_s_private_tmp_pyzmq_release_zmq_d, __pyx_k_private_tmp_pyzmq_release_zmq_d, sizeof(__pyx_k_private_tmp_pyzmq_release_zmq_d), 0, 0, 1, 0},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_r_does_not_provide_a_buffer_int, __pyx_k_r_does_not_provide_a_buffer_int, sizeof(__pyx_k_r_does_not_provide_a_buffer_int), 0, 0, 1, 0},
  {&__pyx_n_s_rc, __pyx_k_rc, sizeof(__pyx_k_rc), 0, 0, 1, 1},
  {&__pyx_n_s_swap_ids, __pyx_k_swap_ids, sizeof(__pyx_k_swap_ids), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_typecode, __pyx_k_typecode, sizeof(__pyx_k_typecode), 0, 0, 1, 1},
  {&__pyx_n_s_zmq, __pyx_k_zmq, sizeof(__pyx_k_zmq), 0, 0, 1, 1},
  {&__pyx_n_s_zmq_devices_monitoredqueue, __pyx_k_zmq_devices_monitoredqueue, sizeof(__pyx_k_zmq_devices_monitoredqueue), 0, 0, 1, 1},
  {&__pyx_n_s_zmq_error, __pyx_k_zmq_error, sizeof(__pyx_k_zmq_error), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 151, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(1, 242, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "buffers.pxd":194
 * 
 *     if PY_MAJOR_VERSION >= 3 and bfmt is not None:
 *         return bfmt.decode('ascii')             # <<<<<<<<<<<<<<
 *     return bfmt
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_n_s_ascii); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "buffers.pxd":242
 *             return PyBuffer_FromReadWriteMemory(ptr, s)
 *     else:
 *         raise NotImplementedError("Old style buffers not available.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Old_style_buffers_not_available); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "zmq/devices/monitoredqueue.pyx":41
 * 
 * 
 * def monitored_queue(Socket in_socket, Socket out_socket, Socket mon_socket,             # <<<<<<<<<<<<<<
 *                     bytes in_prefix=b'in', bytes out_prefix=b'out'):
 *     """monitored_queue(in_socket, out_socket, mon_socket,
 */
  __pyx_tuple__4 = PyTuple_Pack(14, __pyx_n_s_in_socket, __pyx_n_s_out_socket, __pyx_n_s_mon_socket, __pyx_n_s_in_prefix, __pyx_n_s_out_prefix, __pyx_n_s_ins, __pyx_n_s_outs, __pyx_n_s_mons, __pyx_n_s_in_msg, __pyx_n_s_out_msg, __pyx_n_s_swap_ids, __pyx_n_s_msg_c, __pyx_n_s_msg_c_len, __pyx_n_s_rc); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(5, 0, 14, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_private_tmp_pyzmq_release_zmq_d, __pyx_n_s_monitored_queue, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initmonitoredqueue(void); /*proto*/
PyMODINIT_FUNC initmonitoredqueue(void)
#else
PyMODINIT_FUNC PyInit_monitoredqueue(void); /*proto*/
PyMODINIT_FUNC PyInit_monitoredqueue(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_monitoredqueue(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("monitoredqueue", __pyx_methods, __pyx_k_MonitoredQueue_classes_and_funct, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_zmq__devices__monitoredqueue) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "zmq.devices.monitoredqueue")) {
      if (unlikely(PyDict_SetItemString(modules, "zmq.devices.monitoredqueue", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_3zmq_7backend_6cython_7context_Context = __Pyx_ImportType("zmq.backend.cython.context", "Context", sizeof(struct __pyx_obj_3zmq_7backend_6cython_7context_Context), 1); if (unlikely(!__pyx_ptype_3zmq_7backend_6cython_7context_Context)) __PYX_ERR(3, 26, __pyx_L1_error)
  __pyx_vtabptr_3zmq_7backend_6cython_7context_Context = (struct __pyx_vtabstruct_3zmq_7backend_6cython_7context_Context*)__Pyx_GetVtable(__pyx_ptype_3zmq_7backend_6cython_7context_Context->tp_dict); if (unlikely(!__pyx_vtabptr_3zmq_7backend_6cython_7context_Context)) __PYX_ERR(3, 26, __pyx_L1_error)
  __pyx_ptype_3zmq_7backend_6cython_6socket_Socket = __Pyx_ImportType("zmq.backend.cython.socket", "Socket", sizeof(struct __pyx_obj_3zmq_7backend_6cython_6socket_Socket), 1); if (unlikely(!__pyx_ptype_3zmq_7backend_6cython_6socket_Socket)) __PYX_ERR(4, 33, __pyx_L1_error)
  __pyx_vtabptr_3zmq_7backend_6cython_6socket_Socket = (struct __pyx_vtabstruct_3zmq_7backend_6cython_6socket_Socket*)__Pyx_GetVtable(__pyx_ptype_3zmq_7backend_6cython_6socket_Socket->tp_dict); if (unlikely(!__pyx_vtabptr_3zmq_7backend_6cython_6socket_Socket)) __PYX_ERR(4, 33, __pyx_L1_error)
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(5, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(6, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(7, 15, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "zmq/devices/monitoredqueue.pyx":33
 * from zmq.backend.cython.checkrc cimport _check_rc
 * 
 * from zmq import ROUTER, ZMQError             # <<<<<<<<<<<<<<
 * from zmq.error import InterruptedSystemCall
 * 
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ROUTER);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ROUTER);
  __Pyx_INCREF(__pyx_n_s_ZMQError);
  __Pyx_GIVEREF(__pyx_n_s_ZMQError);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_ZMQError);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_zmq, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ROUTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ZMQError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ZMQError, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "zmq/devices/monitoredqueue.pyx":34
 * 
 * from zmq import ROUTER, ZMQError
 * from zmq.error import InterruptedSystemCall             # <<<<<<<<<<<<<<
 * 
 * #-----------------------------------------------------------------------------
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_InterruptedSystemCall);
  __Pyx_GIVEREF(__pyx_n_s_InterruptedSystemCall);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_InterruptedSystemCall);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_zmq_error, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_InterruptedSystemCall); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InterruptedSystemCall, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/devices/monitoredqueue.pyx":41
 * 
 * 
 * def monitored_queue(Socket in_socket, Socket out_socket, Socket mon_socket,             # <<<<<<<<<<<<<<
 *                     bytes in_prefix=b'in', bytes out_prefix=b'out'):
 *     """monitored_queue(in_socket, out_socket, mon_socket,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_3zmq_7devices_14monitoredqueue_1monitored_queue, NULL, __pyx_n_s_zmq_devices_monitoredqueue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_monitored_queue, __pyx_t_1) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/devices/monitoredqueue.pyx":110
 *     return rc
 * 
 * __all__ = ['monitored_queue']             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_monitored_queue);
  __Pyx_GIVEREF(__pyx_n_s_monitored_queue);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_monitored_queue);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_1) < 0) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/devices/monitoredqueue.pyx":1
 * """MonitoredQueue classes and functions.             # <<<<<<<<<<<<<<
 * 
 * Authors
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/checkrc.pxd":6
 * 
 * 
 * cdef inline int _check_rc(int rc) except -1:             # <<<<<<<<<<<<<<
 *     """internal utility for checking zmq return condition
 * 
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init zmq.devices.monitoredqueue", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init zmq.devices.monitoredqueue");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* SaveResetException */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
    #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* Import */
      static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
      static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectCallMethO */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetVTable */
        static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
        #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
        #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
